# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/LfsrConv.pyx
 -------------------------------------------------- part of the fastmat package

  Circular Convolution matrix for sequence generated by Linear Feedback Shift
  Register (LFSR) with certain configuration.


  Author      : wcw
  Introduced  : 2017-07-24
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from .Matrix cimport Matrix
from .Hadamard cimport Hadamard
from .core.types cimport *
from .core.cmath cimport *
from .core.strides cimport *

cdef inline np.int8_t lfsrOutBit(lfsrReg_t state) nogil:
    return (-1 if state & 1 else 1)

cdef inline lfsrReg_t lfsrGenStep(lfsrReg_t state, lfsrReg_t taps,
                                  lfsrReg_t mask) nogil:
    cdef lfsrReg_t tmp = state & taps
    tmp ^= tmp >> 1
    tmp ^= tmp >> 2
    tmp = (tmp & 0x11111111) * 0x11111111
    if tmp & 0x10000000:
        state |= mask

    return state >> 1

cdef inline lfsrReg_t lfsrTapStep(lfsrReg_t state, lfsrReg_t taps,
                                  lfsrReg_t mask) nogil:
    state = state << 1
    if (state & mask) != 0:
        state = state ^ (taps | mask)

    return state

################################################################################
################################################## class Hadamard
cdef class LFSRCirculant(Matrix):

    ############################################## class properties
    # vecC - Property (read-only)
    # Return the sequence defining the circular convolution.
    property vecC:
        def __get__(self):
            return (self._getVecC() if self._vecC is None
                    else self._vecC)

    ############################################## class properties
    # states - Property (read-only)
    # Return the internal register states during the sequence.
    property states:
        def __get__(self):
            return (self._getStates() if self._states is None
                    else self._states)

    # size - Property (read-only)
    # Return the length of the sequence generated
    property size:
        def __get__(self):
            return self._regSize

    # taps - Property (read-only)
    # Return the tap positions of the register
    property taps:
        def __get__(self):
            return (self._regTaps)

    # resetState - Property (read-only)
    # Return the reset state of the LFSR
    property resetState:
        def __get__(self):
            return (self._resetState)

    ############################################## class methods
    def __init__(self, int regSize, lfsrReg_t taps,
                 lfsrReg_t resetState=0xFFFFFFFF, lfsrReg_t length=0):
        '''Initialize Matrix instance'''

        cdef lfsrReg_t mask = 1 << regSize

        # determine register size (determines order of embedded Hadamard)
        self._regTaps = taps
        self._regSize = regSize
        self._resetState = resetState & (mask - 1)

        if self._regSize > 31 or self._regSize < 1:
            raise ValueError("Register sizes only supported from 1 to 31.")

        if self._regTaps >= mask:
            raise ValueError("Tap positions exceeding register size.")

        if self._regTaps == 0:
            raise ValueError("Register must feature at least one feedback tap.")

        if self._resetState == 0:
            raise ValueError("Register reset state must be non-zero.")

        assert length < mask, "Sequence length exceeded register capacity."

        # generate embedded Hadamard matrix
        self._content = (Hadamard(self._regSize), )

        # determine matrix size (square) from the specified amount of steps to
        # take. If no count was given, determine it from the maximal register
        # cycle length for the given polynomial and start value by construction
        # of the determining sequence vector
        cdef lfsrReg_t state = self._resetState
        if length == 0:
            state = lfsrGenStep(self._resetState, taps, mask)
            length = 1

            while state != self.resetState:
                state = lfsrGenStep(state, taps, mask)
                length += 1
                if length >= mask or state == 0:
                    raise ValueError(
                        "Register configuration produces invalid sequence.")

        if length <= 1:
            raise ValueError("Register produces static output.")

        # set properties of matrix
        self._initProperties(length, length, np.int8,
                             cythonCall=True)

    ############################################## class property override
    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self.vecC[(idxN - idxM) % self.numN]

    cpdef np.ndarray _getCol(self, intsize idx):
        '''Return selected columns of self.array'''
        cdef np.ndarray arrRes = _arrEmpty(
            1, self.numN, 0, self._info.dtype[0].typeNum)
        self._roll(arrRes, idx)
        return arrRes

    cpdef np.ndarray _getRow(self, intsize idx):
        '''Return selected rows of self.array'''
        cdef np.ndarray arrRes = _arrEmpty(
            1, self.numN, 0, self._info.dtype[0].typeNum)
        self._roll(arrRes[::-1], self.numN - idx - 1)
        return arrRes

    cpdef Matrix _getNormalized(self):
        return self * np.float32(1. / np.sqrt(self.numN))

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        cdef float complexity = 2 * self._content[0].numN + 2 * self.numN
        return (complexity, complexity)

    ############################################## class core methods
    cdef void _roll(self, np.ndarray vecOut, intsize shift):
        '''Return self.vecC rolled by 'shift' elements.'''
        if shift == 0:
            vecOut[:] = self.vecC
        else:
            vecOut[:shift] = self.vecC[self.numN - shift:]
            vecOut[shift:] = self.vecC[:self.numN - shift]

    cdef np.ndarray _getStates(self):
        cdef intsize nn
        cdef np.ndarray arrStates
        cdef lfsrReg_t[:] mvStates

        cdef nptype typeStates = np.dtype(np.uint32).type_num
        cdef lfsrReg_t mask = 1 << self._regSize
        cdef lfsrReg_t taps = self._regTaps
        cdef lfsrReg_t state = self._resetState

        arrStates = _arrEmpty(1, self.numN, 1, typeStates)
        mvStates = arrStates
        for nn in range(self.numN):
            mvStates[nn] = state & (mask - 1)
            state = lfsrGenStep(state, taps, mask)

        self._states = arrStates
        return arrStates

    cdef np.ndarray _getVecC(self):
        cdef lfsrReg_t mask = 1 << self._regSize
        cdef lfsrReg_t taps = self._regTaps
        cdef lfsrReg_t state = self._resetState
        cdef intsize nn

        cdef np.ndarray arrRes
        cdef np.int8_t[:] mvRes
        cdef nptype typeElement = np.dtype(np.int8).type_num

        arrRes = _arrEmpty(1, self.numN, 1, typeElement)
        mvRes = arrRes
        for nn in range(self.numN):
            mvRes[nn] = lfsrOutBit(state)
            state = lfsrGenStep(state, taps, mask)

        self._vecC = arrRes
        return self._vecC

    cdef void _core(self, np.ndarray arrIn, np.ndarray arrOut,
                    bint flipIn, bint flipOut):

        cdef intsize ii, nn, N = arrIn.shape[0], M = arrIn.shape[1]
        cdef lfsrReg_t state, mask  = 1 << self._regSize
        cdef lfsrReg_t taps         = self._regTaps
        cdef np.ndarray arrData
        cdef STRIDE_s strIn, strData, strOut

        # initialize data Array, no zero init required for maximum length seqs
        if N == mask - 1:
            arrData = _arrEmpty(2, mask, M, _getNpType(arrIn), False)
        else:
            arrData = _arrZero(2, mask, M, _getNpType(arrIn), False)

        strideInit(&strIn, arrIn, 1)
        strideInit(&strOut, arrOut, 1)
        strideInit(&strData, arrData, 1)

        # ensure the first element of each vector in arrData is zero
        opZeroVector(&strData, 0)

        # flipping the arrays achieves circulant shape without actually shifting
        # anything (the original algorithm computes deconvolution which builds
        # an Hankel-like matrix with the sequence elements along anti-diagonals)
        # For the conversion to take place flip the input array during the
        # forward, and the output array during the backward transform.
        # Note, that the flip can be efficiently realized by twiddling around
        # with row indexing

        # apply input permutation P(c->g) from arrIn to arrData (through slices)
        # reset state of address register G resembles sequence start point
        state = self._resetState

        # copy first element from each imput vector to Hadamard transform input
        # data position denoted by the register's start value (copy row to row)
        # then, copy remaining elements (flip them if requested by flipIn)
        # don't flip the first item (which we will we copy explicitly now)
        opCopyVector(&strData, state, &strIn, 0)
        strideSliceVectors(&strIn, 1, -1, 1)        # equivalent to strIn[1:]

        # and now for something completely different: addressing hack!
        # to flip the input we can fiddle around with the array base pointer
        # and its strides. Inverting the slice-stride (keeping row distance but
        # changing direction of traversal) and setting the base pointer to the
        # last element in the array achieves what we want.
        # However, index 0 will be mapped directly to the can as it literally
        # starts after the array data. After that however indexing begins with
        # the last row for index 1.
        if flipIn:
            strideFlipVectors(&strIn)

        for nn in range(0, N - 1):
            state = lfsrGenStep(state, taps, mask)
            opCopyVector(&strData, state, &strIn, nn)

        # apply hadamard-walsh transform from arrData to arrData
        arrData = self._content[0].forward(arrData)

        # reinit slice as arrData has changed
        strideInit(&strData, arrData, 1)

        # apply output permutation P(r->t) fromm arrData to arrOut (slices!)
        # reset state for address register T must always be 1 (unclear in paper)
        state = 1

        # copy elements from Hadamard transform output denoted by the Tap
        # register's start value to first element of result (copy row to row)
        # then, copy remaining elements (flip them if requested by flipOut)
        # don't flip the first item (which we will we copy explicitly now)
        opCopyVector(&strOut, 0, &strData, state)
        strideSliceVectors(&strOut, 1, -1, 1)        # equivalent to strOut[1:]

        # same addressing hack as above
        if flipOut:
            strideFlipVectors(&strOut)

        for nn in range(0, N - 1):
            state = lfsrTapStep(state, taps, mask)
            opCopyVector(&strOut, nn, &strData, state)

    ############################################## class forward / backward
    cpdef _forwardC(self, np.ndarray arrX, np.ndarray arrRes,
                    ftype typeX, ftype typeRes):
        '''
        Calculate the forward transform of this matrix.
        '''
        # dispatch input ndarray to type specialization
        self._core(arrX, arrRes, True, False)

    cpdef _backwardC(self, np.ndarray arrX, np.ndarray arrRes,
                     ftype typeX, ftype typeRes):
        '''
        Calculate the backward transform of this matrix.
        '''
        # dispatch input ndarray to type specialization
        self._core(arrX, arrRes, False, True)

    ############################################### class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        cdef np.ndarray arrRes, vecSequence
        cdef np.int8_t[:] mvSequence
        cdef int ii, state, taps, mask, tmp, cnt

        state = self.resetState
        taps = self.taps
        mask = 1 << self.size
        vecSequence = np.empty((self.numN, ), dtype=np.int8)
        mvSequence = vecSequence
        for ii in range(self.numN):
            mvSequence[ii] = (state & 1) * -2 + 1

            tmp = state & taps
            cnt = 0
            while tmp:
                cnt ^= 1
                tmp &= tmp - 1

            if cnt:
                state |= mask

            state >>= 1

        arrRes = np.empty((self.numN, self.numN), dtype=self.dtype)
        for ii in range(self.numN):
            arrRes[:, ii] = np.roll(vecSequence, ii)

        return arrRes

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                # define matrix sizes and parameters
                'size'          : 4,
                'taps'          : TEST.Permutation([0x9, 0x7]),
                'start'         : TEST.Permutation([0x5, 0x1]),
                'length'        : (lambda param: (15 if param['taps'] == 0x9
                                                  else 7)),
                TEST.NUM_N      : 'length',
                TEST.NUM_M      : TEST.NUM_N,

                # define constructor for test instances and naming of test
                TEST.OBJECT     : LFSRCirculant,
                TEST.INITARGS   : ['size', 'taps', 'start'],
                TEST.NAMINGARGS : dynFormat("%d:%x=%x", 'size', 'taps', 'start')
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH

        # specify tap configurations for various maximum-length-sequences
        db = {1: [0x1],         2: [0x3],           3: [0x3],
              4: [0x9],         5: [0x05],          6: [0x21],
              7: [0x09],        8: [0xC3],          9: [0x011],
              10: [0x081],      11: [0x303],        12: [0xC11],
              13: [0x0C41],     14: [0x1803],       15: [0x4001],
              16: [0xA011],     17: [0x04001],      18: [0x00801],
              19: [0x64001],    20: [0x20001],      21: [0x080001],
              22: [0x200001],   23: [0x040001],     24: [0xC20001],
              25: [0x0400001],  26: [0x3100001],    27: [0x6400001],
              28: [0x2000001],  29: [0x08000001],   30: [0x30000081],
              31: [0x10000001]}

        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN  : (lambda c: LFSRCirculant(c, *db[c])),
                BENCH.FUNC_SIZE : (lambda c: (2 ** c) - 1),
                BENCH.FUNC_STEP : (lambda c: c + 1),
            },
            BENCH.FORWARD: {},
            BENCH.SOLVE: {},
            BENCH.OVERHEAD: {},
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : (lambda c, dt: LFSRCirculant(c, *db[c]))
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Circulant Matrix of Linear Shift Register Sequence ' +
            r'(\texttt{fastmat.LFSRCirculant})',
            DOC.SUBSUBSECTION(
                'Definition and Interface',
                r"""
Linear Feedback Shift Registers (LFSR) as implemented in this class are finite
state machines generating sequences of symbols from the finite field
$F=[-1, +1]$. A shift register of size $N$ is a cascade of $N$ storage elements
$a_n$ for $n = 0,\dots,N-1$, each holding one symbol of $F$. The state of the
shift register is defined by the states of $a_0,\dots,a_{N-1}$.
\cite{lfsr_golomb1967}

The next state of the register is generated from the current state by moving
the contents of each storage element to the next lower index by setting
$a_{n-1} = a_n$ for $n \geq 1$, hence the name shift register. The element
$a_0$ of the current state is discarded completely in the next state. A subset
$T$ of all storage elements with cardinality of 1 or greater is used for
generating the next symbol $a_{N-1}$ by multiplication within $F$. $T$ is
called the tap configuration of the shift register.

The output sequence of the register is the sequence of symbols $a_0$ for each
state of the register. When the shift register repeats one of its previous
states after $L$ state transistions, the output sequence also repeats and
thus is periodic with a length $L$. Evaluation of the sequence starts with
all storage elements set to an initial state $I$. Only periodic sequences of
length $L > 1$ are considered if they also repeat all states including the
initial state and thus form a hamilton circle an the graph corresponding to
the chosen shift register size $N$ and tap configuration $T$.

Instanciation of this matrix class requires supplying the requested register
size $N$, the tap configuration and the initial state. The latter two are
required to be supplied as binary words of up to $N$ bits. A one bit on position
$i$ in the tap configuration adds $a_i$ as \emph{feedback tap} to $T$. At least
one feedback tap must be supplied. The bits in the given initial state word $r$
will be mapped to the initial register state, where $r_n = 0$ sets $a_n = +1$
and $r_n = 1$ sets $a_n = -1$. If no $r$ is given, it is assumed to be
all-ones.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# construct the parameter',
                            'size = 4',
                            'taps = 0b1001',
                            'initial = 0b1010',
                            '',
                            '# construct the matrix',
                            'L = fm.LFSRCirculant(size, taps, initial)',
                            's = L.vecC',
                            caption=r"""
This yields a Circulant matrix where the column-definition vector is the output
of a LFSR of size 4, which is configured to generate a maximum length sequence
of length 15 and a cyclic shift corresponding to the given initial state.
\[\bm s = [+1, -1, +1, -1, -1, +1, +1, -1, +1, +1, +1, -1, -1, -1, -1]\]
\[\bm L = \left(\begin{array}{ccccc}
    +1 &   -1   & -1 &        & -1 \\
    -1 &   +1   & -1 & \dots  & +1 \\
    +1 &   -1   & +1 &        & -1 \\
       & \vdots &    & \ddots &    \\
    -1 &   -1   & -1 &        & +1 \\
\end{array}\right)\]"""),
                r"""
This class depends on \texttt{Hadamard}."""
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on various maximum-length sequences of register
size $k$, where $n = 2^k - 1$.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTSOLVE(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED()
            ),
            DOC.BIBLIO(
                lfsr_golomb1967=DOC.BIBITEM(
                    'Simon W. Golomb',
                    'Shift Register Sequences',
                    'Holden-Day, Inc. 1967')
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/Zero.pyx
 -------------------------------------------------- part of the fastmat package

  All-Zero matrix.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
cimport numpy as np
import numpy as np

from .Matrix cimport Matrix
from .core.cmath cimport _arrZero
from .core.types cimport *

################################################################################
################################################## class Zero
cdef class Zero(Matrix):

    ############################################## class methods
    def __init__(self, numN, numM):
        '''Initialize Matrix instance with its dimensions'''
        # set properties of matrix
        self._initProperties(numN, numM, np.int8)

    ############################################## class property override
    cpdef np.ndarray _getArray(self):
        return _arrZero(2, self.numN, self.numM, self._info.dtype[0].typeNum)

    cpdef np.ndarray _getCol(self, intsize idx):
        return _arrZero(1, self.numN, 1, self._info.dtype[0].typeNum)

    cpdef np.ndarray _getRow(self, intsize idx):
        return _arrZero(1, self.numM, 1, self._info.dtype[0].typeNum)

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return 0

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return 0.

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return 0.

    cpdef Matrix _getT(self):
        return Zero(self.numM, self.numN)

    cpdef Matrix _getH(self):
        return Zero(self.numM, self.numN)

    cpdef Matrix _getConj(self):
        return self

    cpdef Matrix _getGram(self):
        return Zero(self.numM, self.numM)

    cpdef Matrix _getNormalized(self):
        raise ValueError("A Zero Matrix cannot be normalized.")

    ############################################## class property override
    cpdef tuple _getComplexity(self):
        return (self.numN, self.numM)

    ############################################## class forward / backward
    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''
        return _arrZero(
            arrX.ndim, self.numN, arrX.shape[1] if arrX.ndim > 1 else 1,
            _getNpType(arrX))

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''
        return _arrZero(
            arrX.ndim, self.numM, arrX.shape[1] if arrX.ndim > 1 else 1,
            _getNpType(arrX))

    ############################################## class reference
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using
        any fastmat code.
        '''
        return np.zeros((self.numN, self.numM), dtype=self.dtype)

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        return {
            TEST.COMMON: {
                TEST.NUM_N      : 35,
                TEST.NUM_M      : TEST.Permutation([30, TEST.NUM_N]),
                TEST.OBJECT     : Zero,
                TEST.INITARGS   : [TEST.NUM_N, TEST.NUM_M],
                TEST.NAMINGARGS : dynFormat("%dx%d", TEST.NUM_N, TEST.NUM_M)
            },
            TEST.CLASS: {},
            TEST.TRANSFORMS: {}
        }

    def _getBenchmark(self):
        from .inspect import BENCH
        return {
            BENCH.COMMON: {
                BENCH.FUNC_GEN      : (lambda c: Zero(c, c)),
            },
            BENCH.FORWARD: {},
            BENCH.OVERHEAD: {
                BENCH.FUNC_GEN      : (lambda c: Zero(2 ** c, 2 ** c)),
            }
        }

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'Zero Transform (\texttt{fastmat.Zero})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
    \[\bm x \mapsto \bm 0\]
The zero matrix only needs the dimension $n$ of the vectors it acts on. It is
very fast and very good!""",
                DOC.SNIPPET('import fastmat as fm',
                            '',
                            '# define the parameter',
                            'n = 10',
                            '',
                            '# construct the matrix',
                            'O = fm.Zero(n)',
                            caption=r"This yields something trivial.")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm M = \bm 0_{2^k \times 2^k}$; so $n = 2^k$ for $k \in \N$.""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD()
            )
        )



# -*- coding: utf-8 -*-
'''
  demo/sparseReco.py
 -------------------------------------------------- part of the fastmat demos

  Demonstration on how to use fastMat for sparse deconvolution, where one can
  specify most of the parameters involved.


  Author      : sempersn
  Introduced  : 2016-09-28
 ------------------------------------------------------------------------------
  PARAMETERS:
    numSignalSize   - size of the signal and thus also the dictionary
    numSlices       - number of signal to process
    numPulses       - number of pulses in the signal
    numPulseWidth   - width of a single pulse, where 0.5 would be half of
                      the signal length
    numPulseFreq    - frequency relative to the whole length, i.e. number
                      of oszillations over the whole signal

 ------------------------------------------------------------------------------
   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''

import sys
import time

# import numpy functionality
import numpy as np
import numpy.random as npr

################################################## import modules
try:
    import fastmat
except ImportError:
    sys.path.append('..')
    import fastmat


################################################################################
#                          PARAMETER SECTION
################################################################################
# please edit these parameters as you please, but be cautious with the problem
# size, because the traditional method might fail (> 2^13 with 8GB RAM)
################################################################################
numSignalSize = 2**11
numSlices = 1
numPulses = 10
numPulseWidth = 0.01
numPulseFreq = 200


################################################################################
def genPulse(numN, numWidth, numFreq):
    '''
    Construct a pulsed in a gaussian envelope that gets cut off
    to get a signal with compact support, given frequency of the
    cosine as well as the window width.
        numN        - number of samples of the whole pulse
        numWidth    - width of the gaussian window
        numFreq     - frequency of the underlying cosine
        arrT        - array that contains the time values
        arrCos      - Cosine of the time array
        arrGauss    - Envelope of in form of a cut-off Gaussian
        arrP        - the final pulse
    '''

    # create the time array as an
    # open intervall from 0 to 2*pi
    arrT = np.linspace(0, 2 * np.pi * (1. - 1. / numN), numN)

    # construct the cosine of the intervall
    # with the required frequency
    arrCos = np.cos(arrT * numFreq)

    # create a windowed gaussian from
    # -2*sigma to +2*sigma
    arrGauss = np.exp(
        -((arrT - (2 * 2 * np.pi * numWidth)) / (2 * np.pi * numWidth))**2)
    arrGauss[arrT > 4 * 2 * np.pi * numWidth] = 0

    # multiply the two to get the actual pulse
    arrP = arrCos * arrGauss

    # return a normalized version of this pulse
    return arrP / np.sqrt(np.inner(arrP, arrP))


def genGroundTruth(numN, numM, numK):
    '''
    '''

    arrX = np.zeros((numN, numM))

    for ii in range(numM):
        arrInd = int(0.1 * numN) + npr.choice(
            range(int(numN - 0.2 * numN)), numK, replace=False)
        arrX[arrInd, ii] = npr.uniform(1, 2, numK)

    return arrX


################################################################################
#                          CALCULATION SECTION
################################################################################

print("fastmat demo: Sparse reconstruction using efficient convolution")
print("---------------------------------------------------------------")


print(" * Generate the windowed pulse")
s = time.time()
arrP = genPulse(numSignalSize, numPulseWidth, numPulseFreq)
timePulse = time.time() - s


# generate a circulant dictionary which will serve as linear operator for the
# convolution
print(" * Create the dictionary")
s = time.time()
matC = fastmat.Circulant(arrP)
timeDictionary = time.time() - s


# create an explicit version of the dictionary
# that disregards the circulant structure
print(" * Create the unstructured matrix for speed comparison")
matCHat = fastmat.Matrix(matC._getArray())


# generate a random sequence of spikes where position and amplitudes are random
print(" * Generating the ground truth as a sequence of spikes")
s = time.time()
arrX = genGroundTruth(numSignalSize, numSlices, numPulses)
timeGroundTruth = time.time() - s


# do the measurement (apply the forward model), i.e. do the convolution
print(" * Apply the forward model to generate the signal")
s = time.time()
arrY = matC * arrX
timeForward = time.time() - s


# call the sparse recovery algorithm to extract the original pulses
print(" * Do the reconstruction while exploiting structure")
s = time.time()
arrR1 = fastmat.algs.OMP(matC, arrY, numPulses)
timeFastmat = time.time() - s

print(" * Do the reconstruction without exploiting structure")
s = time.time()
arrR2 = fastmat.algs.OMP(matCHat, arrY, numPulses)
timeRegular = time.time() - s

################################################################################
#                               OUTPUT SECTION
################################################################################

print("\nResults:")
print("   Pulse Generation            % 18.3f ms" %(1e3 * timePulse))
print("   Dictionary Generation       % 18.3f ms" %(1e3 * timeDictionary))
print("   Ground Truth Generation     % 18.3f ms" %(1e3 * timeGroundTruth))
print("   Forward Model               % 18.3f ms" %(1e3 * timeForward))
print("")
print("   Reconstruction with fastmat classes % 10.3f s" %(timeFastmat))
print("   Reconstruction with dense matrices  % 10.3f s" %(timeRegular))
print("   Speedup factor              % 18.3f" %(timeRegular / timeFastmat))
print("")
print("   Efficient Matrix Storage    % 18.3f MB" %(matC.nbytes / 2. ** 20))
print("   Reference Matrix Storage    % 18.3f MB" %(matCHat.nbytes / 2. ** 20))




# -*- coding: utf-8 -*-
'''
  fastmat/Matrix.pyx
 -------------------------------------------------- part of the fastmat package

  Base class for all fast matrix transformations in fastmat.


  Author      : wcw, sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
TODO:
    - add caches for cols / rows
    - implement generation of normalized matrix of general matrices
'''
from libc.string cimport memset
from libc.math cimport isnan

import types
import sys

import numpy as np
cimport numpy as np
from scipy.sparse import spmatrix

# initialize numpy C-API interface for cython extension-type classes.
# This call is required once for every module that uses PyArray_ calls
# WARNING: DO NOT REMOVE THIS LINE OR SEGFAULTS ARE ABOUT TO HAPPEN!
np.import_array()

from .core.types cimport *
from .core.cmath cimport *
from .core.calibration import getMatrixCalibration
#_conjugate, _arrEmpty, _arrReshape, _arrForceContType
from .Product cimport Product
from .Sum cimport Sum
from .Diag cimport Diag

# have a very lazy import to avoid various package imports during main init
getMemoryFootprint = None

################################################################################
################################################## class FastmatFlags
cdef class FastmatFlags(object):
    def __init__(self):
        self.bypassAllow = True
        self.bypassAutoArray = True

flags = FastmatFlags()

################################################################################
################################################## class MatrixCalibration
cdef class MatrixCalibration(object):
    def __init__(self, offsetForward, offsetBackward,
                 gainForward, gainBackward):
        self.gainForward = gainForward
        self.gainBackward = gainBackward
        self.offsetForward = offsetForward
        self.offsetBackward = offsetBackward

    def export(self):
        return (self.offsetForward, self.offsetBackward,
                self.gainForward, self.gainBackward)

    def __repr__(self):
        return str(self.export())


cdef tuple profileToTuple(PROFILE_s profile):
    return (profile.overhead,
            profile.effort,
            profile.overheadNested,
            profile.effortNested,
            profile.complexity)

cdef void finishProfile(PROFILE_s *profile, float offset, float gain):
    profile[0].overhead = profile[0].overheadNested + offset
    profile[0].effort   = profile[0].effortNested + gain * profile[0].complexity

cdef bint profileUpdate(PROFILE_s *profile, intsize numM, bint allowBypass,
                        PROFILE_s *profClass, PROFILE_s *profBypass):
    cdef float estimateClass, estimateBypass
    cdef PROFILE_s *profSelected

    estimateClass  = profClass[0].overhead + profClass[0].effort * numM
    estimateBypass = profBypass[0].overhead + profBypass[0].effort * numM

    profSelected = (profBypass if (allowBypass and
                                   estimateBypass < estimateClass)
                    else profClass)
    profile[0].overheadNested += profSelected[0].overhead
    profile[0].effortNested   += profSelected[0].effort * numM

    return profSelected == profClass

################################################################################
################################################## class Matrix
cdef class Matrix(object):
    '''
    class Matrix of fastmat

    Matrix is the base class of which all fastmat matrices are derived from.
    Many routines, which are available for all classes, are defined here.
    Please note: Many properties defined here link to private variables which
    are overwritten in child classes.
        e.g. self.numN --> self._numN, which may be overwritten in
        child.__init__() to reflect childs' dimension.
    '''

    ############################################## basic class properties
    # dtype - Property (read-only)
    # Report the matrix elements' data type
    property dtype:
        def __get__(self):
            return np.PyArray_TypeObjectFromType(self._info.dtype.typeNum)

    # dtypeNum - Property (read-only)
    # Report the matrix elements' data type (fastmat fused type number)
    property dtypeNum:
        def __get__(self):
            return self._info.dtype[0].fusedType

    # numN - Property (read-only)
    # Report the matrix row count
    property numN:
        def __get__(self):
            return self._info.numN

    # numM - Property (read-only)
    # Report the matrix column count
    property numM:
        def __get__(self):
            return self._info.numM

    # shape - Property (read-only)
    # Report the matrix shape as tuple of (rows, columns)
    property shape:
        def __get__(self):
            return (self.numN, self.numM)

    # tag - Property (read-write)
    # Description tag of Matrix
    property tag:
        def __get__(self):
            return (self._tag)

        def __set__(self, tag):
            self._tag = tag

    # bypassAllow - Property (read-write)
    # enable or disable transform bypassing based on performance estimates
    property bypassAllow:
        def __get__(self):
            return self._bypassAllow

        def __set__(self, value):
            self._bypassAllow = value

    # bypassAutoArray - Property (read-write)
    # enable or disable automatic dense array generation for transform bypass
    property bypassAutoArray:
        def __get__(self):
            return self._bypassAutoArray

        def __set__(self, value):
            self._bypassAutoArray = value

    ############################################## class resource handling
    # nbytes - Property(read)
    # Size of the Matrix object
    property nbytes:
        def __get__(self):
            global getMemoryFootprint
            if getMemoryFootprint is None:
                getMemoryFootprint = __import__(
                    'fastmat.core.resource', globals(), locals(),
                    ['getMemoryFootprint']).getMemoryFootprint

            return getMemoryFootprint(self)

    # _nbytesReference - Property(read)
    # Size of the Matrix object
    property nbytesReference:
        def __get__(self):
            if self._forwardReferenceMatrix is None:
                self._forwardReferenceInit()

            global getMemoryFootprint
            if getMemoryFootprint is None:
                getMemoryFootprint = __import__(
                    '.core.resource', globals(), locals(),
                    ['getMemoryFootprint']).getMemoryFootprint

            return getMemoryFootprint(self._forwardReferenceMatrix)

    def __copy__(self):
        '''
        Performs a copy() on this class instance. As fastmat matrix classes are
        defined immutable nothing must be copied, so retuning self suffices.
        '''
        return self

    def __deepcopy__(self, dict memo):
        '''
        Performs a deepcopy() on this class instance. As fastmat matrix classes
        are defined immutable nothing must be copied, so retuning self suffices.
        '''
        return self

    ############################################## array, col, row & item access
    # array - Property (read-only)
    # Return a dense array representation of the matrix
    property array:
        def __get__(self):
            return (self.getArray() if self._array is None else self._array)

    def __array__(self):
        '''
        Return an expanded array representation of matrix. Invoked internally
        during broadcasting, redirect to self._getArray().
        '''
        return self._getArray()

    cpdef np.ndarray getArray(self):
        if self.__class__ is not Matrix:
            self._array = self._getArray()

        return self._array

    cpdef np.ndarray _getArray(self):
        '''
        Return an explicit representation of the matrix as numpy-array.
        '''
        # check whether this instance is a Matrix or some child class
        if self.__class__ == Matrix:
            # it is a Matrix, so return our internal data
            return self._array
        else:
            # it is any child class. Use self.forward() to generate a ndarray
            # representation of this matrix. This calculation is very slow but
            # also very general. Therefore it is used also in child classes
            # when no explicit code for self._getArray() is provided.
            return self._forward(np.eye(self.numM, dtype=self.dtype))

    def getCols(self, indices):
        '''Return selected columns of self.getArray()'''
        cdef np.ndarray arrResult, arrIdx
        cdef intsize ii, numSize

        if np.isscalar(indices):
            arrResult = self.getCol(indices)
        else:
            arrIdx = np.array(indices)
            if arrIdx.ndim > 1:
                raise ValueError("Index array must have at most one dimension.")

            # if a dense representation already exists, use it!
            if self._array is not None:
                arrResult = self._array[:, indices]
            else:
                numSize = arrIdx.size
                arrResult = _arrEmpty(2, self._info.numN, numSize,
                                      self._info.dtype[0].typeNum)

                for ii in range(numSize):
                    arrResult[:, ii] = self.getCol(arrIdx[ii])

        return arrResult

    def getCol(self, idx):
        # TODO it might be nice to offer some cache for the columns here
        if idx < 0 or idx >= self.numM:
            raise ValueError("Column index exceeds matrix dimensions.")

        # if a dense representation already exists, use it!
        if self._array is not None:
            return np.squeeze(self._array[:, idx])

        return self._getCol(idx)

    cpdef np.ndarray _getCol(self, intsize idx):
        cdef np.ndarray arrData = _arrZero(1, self._info.numM, 1,
                                           self._info.dtype[0].typeNum)
        arrData[idx] = 1
        return self.forward(arrData)

    def getRows(self, indices):
        '''Return selected rows of self.getArray()'''
        cdef np.ndarray arrResult, arrIdx
        cdef intsize ii, numSize

        if np.isscalar(indices):
            arrResult = self.getRow(indices)
        else:
            arrIdx = np.array(indices)
            if arrIdx.ndim > 1:
                raise ValueError("Index array must have at most one dimension.")

            # if a dense representation already exists, use it!
            if self._array is not None:
                arrResult = self._array[indices, :]
            else:
                numSize = arrIdx.size
                arrResult = _arrEmpty(2, numSize, self._info.numM,
                                      self._info.dtype[0].typeNum)

                for ii in range(numSize):
                    arrResult[ii, :] = self.getRow(arrIdx[ii])

        return arrResult

    def getRow(self, idx):
        # TODO it might be nice to offer some cache for the columns here
        if idx < 0 or idx >= self.numN:
            raise ValueError("Row index exceeds matrix dimensions.")

        # if a dense representation already exists, use it!
        if self._array is not None:
            return np.squeeze(self._array[idx, :])

        return self._getRow(idx)

    cpdef np.ndarray _getRow(self, intsize idx):
        cdef np.ndarray arrData = _arrZero(1, self._info.numN, 1,
                                           self._info.dtype[0].typeNum)
        arrData[idx] = 1
        return self.backward(arrData).conj()

    def __getitem__(self, tplIdx):
        '''
        Return the element at index tplIdx. The element is evaluated using
        self.forward() on a column-selecting vector, which is then indexed at
        the row-index to retrieve the element. This is quite slow and memory
        consuming for large matrices due to the use of numpy.eye().
        '''
        if len(tplIdx) != 2:
            raise ValueError("Matrix element access requires two indices.")

        cdef intsize idxN = tplIdx[0], idxM = tplIdx[1]
        cdef intsize N = self._info.numN, M = self._info.numM

        if idxN < 0 or idxN >= N or idxM < 0 or idxM >= M:
            raise IndexError("Index %s exceeds matrix dimensions %s." %(
                str(tplIdx), str(self.shape)))

        # if a dense representation already exists, use it!
        if self._array is not None:
            return self._array[idxN, idxM]

        return self._getItem(idxN, idxM)

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self._getCol(idxM)[idxN]

    ############################################## Matrix content property
    # content - Property (read-only)
    # Return the Matrix contents
    property content:
        def __get__(self):
            return self._content

    ############################################## algorithmic properties
    # largestEV - Property (read-only)
    # Return the largest eigenvalue for this matrix instance
    property largestEV:
        def __get__(self):
            return (self.getLargestEV() if self._largestEV is None
                    else self._largestEV)

    def getLargestEV(self, maxSteps=10000,
                     relEps=1., eps=0., alwaysReturn=False):
        if self._numN != self._numM:
            raise ValueError("largestEV: Matrix must be square.")

        result = self._getLargestEV(maxSteps, relEps, eps, alwaysReturn)
        self._largestEV = self._largestEV if np.isnan(result) else result
        return result

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        '''
        Determine largest eigen-value of a fastmat matrix.

        The following variables are used:
        eps            - relative stopping threshold
        numNormOld     - norm of vector from last iteration
        numNormNew     - norm of vector in current iteration
        vecBNew        - normalized current iterate
        vecBOld        - normalized last iterate
        '''
        cdef np.ndarray vecBOld, vecBNew, vecBNewHat

        # determine an convergance threshold if eps is deliberately set to zero
        if eps == 0:
            eps = relEps * _getTypeEps(safeTypeExpansion(self.dtype)) * \
                (self.numN if self.numN >= self.numM else self.numM)

        if self.numN != self.numM:
            raise ValueError("largestEV: Matrix must be square.")

        # sample one point uniformly in space, have a zero-vector reference
        vecBNew = np.random.randn(self.numM).astype(
            np.promote_types(np.float32, self.dtype))
        vecBNew /= np.linalg.norm(vecBNew)
        vecBOld = np.zeros((<object> vecBNew).shape, vecBNew.dtype)

        # now continiously apply the matrix and renormalize until convergence
        for numSteps in range(maxSteps):
            vecBOld = vecBNew
            vecBNew = self.forward(vecBOld)

            normNew = np.linalg.norm(vecBNew)
            if normNew == 0:
                # presumably a zero-matrix
                return vecBNew.dtype.type(0.)

            vecBNew /= normNew

            if np.linalg.norm(vecBNew - vecBOld) < eps:
                vecBNewHat = vecBNew.conj()
                return (np.inner(vecBNewHat, self.forward(vecBNew)) /
                        np.inner(vecBNewHat, vecBNew))

        # did not converge - return NaN
        if alwaysReturn:
            vecBNewHat = vecBNew.conj()
            return (np.inner(vecBNewHat, self.forward(vecBNew)) /
                    np.inner(vecBNewHat, vecBNew))
        else:
            return vecBNew.dtype.type(np.NaN)

    # largestSV - Property (read-only)
    # Return the largestSV for this matrix instance
    property largestSV:
        def __get__(self):
            return (self.getLargestSV() if self._largestSV is None
                    else self._largestSV)

    def getLargestSV(self, maxSteps=10000,
                     relEps=1., eps=0., alwaysReturn=False):
        result = self._getLargestSV(maxSteps, relEps, eps, alwaysReturn)
        self._largestSV = self._largestSV if np.isnan(result) else result
        return result

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        '''
        Determine largest singular value of a fastmat matrix.

        The following variables are used:
        eps            - relative stopping threshold
        vecBNew        - normalized current iterate
        vecBOld        - normalized last iterate
        '''
        cdef np.ndarray vecBOld, vecBNew
        cdef Matrix matGram = self.gram
        cdef intsize ii

        # determine an convergance threshold if eps is deliberately set to zero
        if eps == 0:
            eps = relEps * _getTypeEps(safeTypeExpansion(self.dtype)) * \
                (self.numN if self.numN >= self.numM else self.numM)

        # sample one initial sample, have a zero-vector reference
        vecB = np.random.randn(self.numM).astype(
            np.promote_types(np.float64, self.dtype))
        normNew = np.linalg.norm(vecB)

        # iterate until changes cool down
        for ii in range(maxSteps):
            vecB = matGram.forward(vecB / normNew)

            normOld = normNew
            normNew = np.linalg.norm(vecB)
            if normNew == 0:
                # presumably a zero-matrix
                return vecB.dtype.type(0.)

            if np.abs(normNew - normOld) < eps:
                # return square root of the current norm after applying the gram
                # matrix
                return np.sqrt(normNew)

        # did not converge - return NaN
        return (np.sqrt(normNew) if alwaysReturn else np.float64(np.NaN))

    ############################################## generic algebric properties
    # gram - Property (read-only)
    # Return the gram matrix for this fastmat class
    property gram:
        def __get__(self):
            return (self.getGram() if self._gram is None
                    else self._gram)

    def getGram(self):
        self._gram = self._getGram()
        return self._gram

    cpdef Matrix _getGram(self):
        return Product(self.H, self)

    # normalized - Property (read-only)
    # Return a normalized matrix for this instance
    property normalized:
        def __get__(self):
            return (self.getNormalized() if self._normalized is None
                    else self._normalized)

    def getNormalized(self):
        self._normalized = self._getNormalized()
        return self._normalized

    cpdef Matrix _getNormalized(self):
        # determine type of normalization diagonal matrix
        diagType = safeTypeExpansion(self.dtype)

        # array that contains the norms of each column
        cdef np.ndarray arrDiag = np.empty(self.numM, dtype=diagType)

        # number of elements we consider at once during normalization
        cdef intsize numStrideSize = 256

        cdef np.ndarray arrSelector = np.zeros(
            (self.numM, min(numStrideSize, self.numM)), dtype=diagType)

        cdef int ii
        for ii in range(0, self.numM, numStrideSize):
            vecIndices = np.arange(ii, min(ii + numStrideSize, self.numM))
            if ii == 0 or vecSlice.size != vecIndices.size:
                vecSlice = np.arange(0, vecIndices.size)

            arrSelector[vecIndices, vecSlice] = 1

            arrDiag[vecIndices] = np.linalg.norm(
                self.forward(
                    arrSelector if vecIndices.size == arrSelector.shape[1]
                    else arrSelector[:, vecSlice]),
                axis=0)
            arrSelector[vecIndices, vecSlice] = 0

        # check if we've found any zero
        if np.any(arrDiag == 0):
            raise ValueError("Normalization: Matrix has zero-norm column.")

        # finally invert the diagonal and generate normalized matrix
        return self * Diag(1. / arrDiag)

    # T - Property (read-only)
    # Return the transpose of the matrix as fastmat class
    property T:
        def __get__(self):
            return (self.getT() if self._T is None else self._T)

    def getT(self):
        self._T = self._getT()
        return self._T

    cpdef Matrix _getT(self):
        return Transpose(self)

    # H - Property (read-only)
    # Return the hermitian transpose of the matrix as fastmat class
    property H:
        def __get__(self):
            return (self.getH() if self._H is None else self._H)

    def getH(self):
        self._H = self._getH()
        return self._H

    cpdef Matrix _getH(self):
        return Hermitian(self)

    # conj - Property (read-only)
    # Return the conjugate of the matrix as fastmat class
    property conj:
        def __get__(self):
            return (self.getConj() if self._conj is None else self._conj)

    def getConj(self):
        self._conj = self._getConj()
        return self._conj

    cpdef Matrix _getConj(self):
        return getConjugate(self)

    ############################################## computation complexity
    # complexityForward - Property (read-only)
    # Return the computational complexity of all functionality implemented in
    # the class itself, not including calls of external code.
    property complexity:
        def __get__(self):
            return (self._profileForward.complexity,
                    self._profileBackward.complexity)

    def getComplexity(self):
        cdef tuple complexity = self._getComplexity()
        assert complexity is not None
        self._profileForward.complexity = complexity[0]
        self._profileBackward.complexity = complexity[1]

    cpdef tuple _getComplexity(self):
        cdef float complexity = self.numN * self.numM
        return (complexity, complexity)

    # profile - Property (read-only)
    # Return the timing profiles of this specific class. These are also used to
    # determine which computation strategy to use for a perticular input.
    # (distinct amount of vectors)
    property profile:
        def __get__(self):
            return (profileToTuple(self._profileForward),
                    profileToTuple(self._profileBackward))

    property profileForward:
        def __get__(self):
            return self._profileForward

    property profileBackward:
        def __get__(self):
            return self._profileBackward

    property profileBypassFwd:
        def __get__(self):
            return self._profileBypassFwd

    property profileBypassBwd:
        def __get__(self):
            return self._profileBypassBwd

    def estimateRuntime(self, intsize M=1):
        cdef float estimateForward  = (self._profileForward.overhead +
                                       self._profileForward.effort * M)
        cdef float estimateBypassFwd = (self._profileBypassFwd.overhead +
                                        self._profileBypassFwd.effort * M)
        cdef float estimateBackward = (self._profileBackward.overhead +
                                       self._profileBackward.effort * M)
        cdef float estimateBypassBwd = (self._profileBypassBwd.overhead +
                                        self._profileBypassBwd.effort * M)
        return (estimateBypassFwd if (self._bypassAllow and
                                      (self._array is not None or
                                       self._bypassAutoArray) and
                                      (estimateBypassFwd < estimateForward))
                else estimateForward,
                estimateBypassBwd if (self._bypassAllow and
                                      (self._arrayH is not None or
                                       self._bypassAutoArray) and
                                      (estimateBypassBwd < estimateBackward))
                else estimateBackward)

    cdef void _initProfiles(self):
        '''
        Generate performance profiles based on intrinsic complexity and
        external dependencies (e.g. nested class calls) and condense results
        in an overall figure of merit based on:
          - computational complexity of the particular class instance
          - calibration parameters of class (map complexity to runtime)
          - nested calls to other fastmat matrix instances

        The resulting figure is divided into a fixed overhead and a variable
        effort per data vector the transform is applied to.

        The overall figures shall become valid (non-zero) if and only if all of
        the following requirements are met:
          - valid calibration parameters could be loaded for this class
          - in case of meta classes all embedded matrix classes must report
            valid (non-zero) values for overall overhead and effort

        Separate profiles will be generated for the
          - forward and backward transforms of this particular class instance
          - a bypass transform based on dot-product complexity
        '''
        # initialize profiles
        memset(&(self._profileForward), 0, sizeof(PROFILE_s))
        memset(&(self._profileBackward), 0, sizeof(PROFILE_s))
        memset(&(self._profileBypassFwd), 0, sizeof(PROFILE_s))
        memset(&(self._profileBypassBwd), 0, sizeof(PROFILE_s))

        # determine complexity of class instance transforms and Bypass transform
        self.getComplexity()
        cdef tuple complexity = Matrix._getComplexity(self)
        self._profileBypassFwd.complexity = complexity[0]
        self._profileBypassBwd.complexity = complexity[1]

        # Explore the performance profiles of nested fastmat classes
        # Fills in the profile fields overheadNested and effortNested
        self._exploreNestedProfiles()

        # compute performance profile of this class
        cdef float nan = np.nan
        cdef MatrixCalibration calClass = getMatrixCalibration(self.__class__)
        cdef MatrixCalibration calBase  = getMatrixCalibration(Matrix)
        finishProfile(&(self._profileForward),
                      calClass.offsetForward if calClass is not None else nan,
                      calClass.gainForward if calClass is not None else nan)
        finishProfile(&(self._profileBackward),
                      calClass.offsetBackward if calClass is not None else nan,
                      calClass.gainBackward if calClass is not None else nan)
        finishProfile(&(self._profileBypassFwd),
                      calBase.offsetForward if calBase is not None else nan,
                      calBase.gainForward if calBase is not None else nan)
        finishProfile(&(self._profileBypassBwd),
                      calBase.offsetBackward if calBase is not None else nan,
                      calBase.gainBackward if calBase is not None else nan)

    cpdef _exploreNestedProfiles(self):
        '''
        Explore the runtime properties of all nested fastmat matrices. Use ane
        iterator on self._content by default to sum the profile properties of
        all nested classes of meta-classes by default. basic-classes either
        have an empty tuple for _content or need to overwrite this method.
        '''
        cdef Matrix item
        cdef bint bypass

        for item in self:
            bypass = (item._bypassAllow and
                      (item._array is not None or item._bypassAutoArray))
            profileUpdate(&(self._profileForward), 1, bypass,
                          &(item._profileForward), &(item._profileBypassFwd))
            profileUpdate(&(self._profileBackward), 1, bypass,
                          &(item._profileBackward), &(item._profileBypassBwd))

    ############################################## class methods
    def __init__(self, arrMatrix):
        '''Initialize Matrix instance with numpy-array from arrMatrix.'''
        if not isinstance(arrMatrix, np.ndarray):
            raise TypeError("Matrix: Use Sparse() for scipy spmatrix."
                            if isinstance(arrMatrix, spmatrix)
                            else "Matrix: Matrix is not a numpy ndarray")

        dims = len(arrMatrix.shape)
        if dims == 2:
            self._array = np.copy(arrMatrix)
        elif dims < 2:
            self._array = np.reshape(np.copy(arrMatrix), (len(arrMatrix), 1))
        else:
            raise NotImplementedError("Matrix data array must be 2D")

        # set properties of matrix
        self._initProperties(
            self._array.shape[0],        # numN
            self._array.shape[1],        # numM
            self._array.dtype            # data type of matrix
        )

    def _initProperties(
        self,
        intsize numN,
        intsize numM,
        object dataType,
        **properties
    ):
        '''
        Initialize properties of the matrix representation.
        Use explicit values for type properties.
        '''

        # assign basic class properties (at c-level)
        self._info.nDim = 2
        self._info.numN = numN
        self._info.numM = numM
        self._info.dtype = _getTypeInfo(dataType)

        # get and assign c-level properties
        self._cythonCall          = properties.pop('cythonCall', False)
        self._forceInputAlignment = properties.pop('forceInputAlignment', False)
        self._widenInputDatatype  = properties.pop('widenInputDatatype', False)
        self._useFortranStyle     = properties.pop('fortranStyle', True)
        self._bypassAllow         = properties.pop('bypassAllow',
                                                   flags.bypassAllow)

        # determine new value of _bypassAutoArray: take the setting of `flags`
        # as default but heck (if True) that no child class has bypassAutoArray
        # set to False. In this case the class AutoArray function will be
        # triggered indirectly if the parent class disregards this setting.
        cdef bint autoArray = (flags.bypassAutoArray and
                               all(not item.bypassAutoArray for item in self))
        self._bypassAutoArray     = properties.pop('bypassAutoArray', autoArray)

        # initialize performance profile
        self._initProfiles()

    def __repr__(self):
        '''
        Return a string representing this very class instance. For
        identification module name, class name, instance id and shape will be
        returned formatted as string.
        '''
        return "<%s[%dx%d]:0x%12x>" %(
            self.__class__.__name__,
            self.numN,
            self.numM,
            id(self)
        )

    def __str__(self):
        '''
        Return a string representing the classes' contents in a more
        human-readable and human-interpretable fashion. Currently the
        call is redirected to self.__repr__().
        '''
        return self.getArray().__str__()

    def __len__(self):
        '''Return number of nested elements in matrix instance.'''
        return 0 if self._content is None else len(self._content)

    def __iter__(self):
        '''Iterate through all nested objects of this matrix instance.'''
        return self if self._content is None else self._content.__iter__()

    def __next__(self):
        '''Stop iteration as __iter__ redirected here. Python3-Style.'''
        raise StopIteration

    def next(self):
        '''Stop iteration as __iter__ redirected here. Python2-Style.'''
        raise StopIteration

    ############################################## class operator overloading
    def __add__(self, element):
        '''Return the sum of this matrix instance and another.'''
        return Sum(self, element)

    def __sub__(self, element):
        '''Return the difference of this matrix instance and another.'''
        return Sum(self, (-1) * element)

    def __mul__(self, factor):
        '''Return the product of this matrix and another or a scalar.'''
        if isinstance(factor, np.ndarray):
            return self.forward(factor)
        else:
            return Product(self, factor)

    def __rmul__(self, factor):
        '''Return the product of a scalar and this matrix.'''
        if np.isscalar(factor) or isinstance(factor, Matrix):
            return Product(factor, self)
        else:
            raise TypeError("Invalid product term for fastmat Matrix.")

    def __div__(self, divisor):
        '''Return the product of a matrix by the reciproce of a given scalar.'''
        if np.isscalar(divisor):
            if divisor != 0:
                return Product(self, 1. / divisor)
            else:
                raise ZeroDivisionError("Division by Zero.")
        else:
            raise TypeError("Only scalars allowed as divisors.")

    def __truediv__(self, divisor):
        return self.__div__(divisor)

    def __floordiv__(self, divisor):
        raise NotImplementedError("Matrix Floor division not allowed.")

    ############################################## class forward / backward
    cpdef _forwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        raise NotImplementedError("No _forwardC method implemented in class.")

    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''
        Perform a forward transform for general matrices. This method will get
        overwritten in child classes of Matrix to implement specific transforms.
        This base function will also be called when a cython-call object does
        not define a _forward() method. One circumstance leading to this is
        when the runtime estimation within the forward()-entry point decides to
        bootstrap a dense array representation from within forward(). Then
        _getArray() cannot simply call forward() as this would leed to an
        infinite loop. Then, _forward() will be called directly, leading to this
        issue with cythonCall classes that only define a _forwardC()
        '''
        cdef nptype typeInput, typeOutput
        if self._cythonCall:
            # Create output array
            typeInput = typeSelection[np.PyArray_TYPE(arrX)]
            typeOutput = typeInfo[typeInput].promote[
                self._info.dtype[0].fusedType]

            arrOutput = _arrEmpty(2, self.numN, arrX.shape[1],
                                  typeInfo[typeOutput].typeNum)
            self._forwardC(arrX, arrOutput, typeInput, typeOutput)
            return arrOutput
        else:
            return self._array.dot(arrX)

    cpdef np.ndarray forward(self, np.ndarray arrX):
        '''
        Calculate the forward transform (self * x).
        Dimension-checking is performed to ensure valid fast transforms as these
        may succeed even when dimensions do not match. To support both single-
        and multidimensional input vectors x, single dimensional input will be
        reshaped to (n, 1) before processing and flattened to (n) after
        completion. This allows the use of both vectors and arrays. The actual
        transform code gets called by the callbacks specified in funcPython and
        funcCython, depending on the state of self._cythonCall.

        !!! Do not override this method !!!
        '''
        # local variable holding return array
        cdef np.ndarray arrInput = arrX, arrOutput
        cdef int ndimInput = arrInput.ndim
        cdef ftype typeInput, typeOutput

        # check dimension count and sizes
        if ndimInput > 2:
            raise ValueError("Input data array must be at most 2D")

        # arrInput.N must match self.M in forward() and .N in backward()
        if arrInput.shape[0] != self.numM:
            raise ValueError("Dimension mismatch %s <-!-> %s" %(
                str(self.shape), str((<object> arrInput).shape)))

        # force array of data to be two-dimensional and determine vector count
        if ndimInput == 1:
            arrInput = _arrReshape(arrInput, 2, self.numM, 1, np.NPY_ANYORDER)
        cdef intsize M = arrInput.shape[1]

        # Determine output data type
        typeInput = typeSelection[np.PyArray_TYPE(arrInput)]
        typeOutput = typeInfo[typeInput].promote[self._info.dtype[0].fusedType]

        # force input data type to fulfill some requirements if needed
        #  * check for data type match
        #  * check for data alignment (contiguousy and segmentation)
        typeForce = (typeInfo[typeOutput].typeNum if self._widenInputDatatype
                     else typeInfo[typeInput].typeNum)
        if self._forceInputAlignment:
            arrInput = _arrForceTypeAlignment(arrInput, typeForce, 0,
                                              self._useFortranStyle)
        else:
            if self._widenInputDatatype:
                arrInput = _arrForceType(arrInput, typeForce)

        # estimate runtimes according profileForward and profileBypass
        # if the dot-product bypass strategy leads to smaller runtimes, do it!
        cdef float estimateForward  = (self._profileForward.overhead +
                                       self._profileForward.effort * M)
        cdef float estimateBypass   = (self._profileBypassFwd.overhead +
                                       self._profileBypassFwd.effort * M)
        if (self._bypassAllow and estimateBypass < estimateForward and
                (self._array is not None or self._bypassAutoArray)):
            arrOutput = self.array.dot(arrInput)
        else:
            # call fast transform with either cython or python style
            if self._cythonCall:
                # Create output array
                arrOutput = _arrEmpty(
                    2, self.numN, M if ndimInput > 1 else 1,
                    typeInfo[typeOutput].typeNum)

                # Call calculation routine (fused type dispatch must be
                # be done locally in each class, typeInfo gets passed)
                self._forwardC(arrInput, arrOutput, typeInput, typeOutput)
            else:
                arrOutput = self._forward(arrInput)

        if ndimInput == 1:
            # reshape back to single-dimensional vector
            arrOutput = _arrReshape(arrOutput, 1, self.numN, 1, np.NPY_ANYORDER)

        return arrOutput

    cpdef _backwardC(
        self,
        np.ndarray arrX,
        np.ndarray arrRes,
        ftype typeX,
        ftype typeRes
    ):
        raise NotImplementedError("No _backwardC method implemented in class.")

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''
        Perform a backward transform for general matrices. This method
        may get overwritten in child classes of Matrix
        '''
        return _conjugate(self._array.T).dot(arrX)

    cpdef np.ndarray backward(self, np.ndarray arrX):
        '''
        Calculate the backward transform (self ^ H * x) where ^H is the
        hermitian transpose operator.
        Dimension-checking is performed to ensure valid fast transforms as these
        may succeed even when dimensions do not match. To support both single-
        and multidimensional input vectors x, single dimensional input will be
        reshaped to (n, 1) before processing and flattened to (n) after
        completion. This allows the use of both vectors and arrays. The actual
        transform code gets called by the callbacks specified in funcPython and
        funcCython, depending on the state of self._cythonCall.

        !!! Do not override this method !!!
        '''
        # local variable holding return array
        cdef np.ndarray arrInput = arrX, arrOutput
        cdef int ndimInput = arrInput.ndim
        cdef ftype typeInput, typeOutput
        cdef nptype typeForce

        # check dimension count and sizes
        if ndimInput > 2:
            raise ValueError("Input data array must be at most 2D")

        # arrInput.N must match self.M in forward() and .N in backward()
        if arrInput.shape[0] != self.numN:
            raise ValueError("Dimension mismatch %s.H <-!-> %s" %(
                str(self.shape), str((<object> arrInput).shape)))

        # force array of data to be two-dimensional and determine vector count
        if ndimInput == 1:
            arrInput = _arrReshape(arrInput, 2, self.numN, 1, np.NPY_ANYORDER)
        cdef intsize M = arrInput.shape[1]

        # Determine output data type
        typeInput = typeSelection[np.PyArray_TYPE(arrInput)]
        typeOutput = typeInfo[typeInput].promote[self._info.dtype[0].fusedType]

        # force input data type to fulfill some requirements if needed
        #  * check for data type match
        #  * check for data alignment (contiguousy and segmentation)
        typeForce = (typeInfo[typeOutput].typeNum if self._widenInputDatatype
                     else typeInfo[typeInput].typeNum)
        if self._forceInputAlignment:
            arrInput = _arrForceTypeAlignment(arrInput, typeForce, 0,
                                              self._useFortranStyle)
        else:
            if self._widenInputDatatype:
                arrInput = _arrForceType(arrInput, typeForce)

        # estimate runtimes according profileForward and profileBypass
        # if the dot-product bypass strategy leads to smaller runtimes, do it!
        cdef float estimateBackward = (self._profileBackward.overhead +
                                       self._profileBackward.effort * M)
        cdef float estimateBypass   = (self._profileBypassBwd.overhead +
                                       self._profileBypassBwd.effort * M)
        if (self._bypassAllow and estimateBypass < estimateBackward and
                (self._arrayH is not None or self._bypassAutoArray)):
            if self._arrayH is None:
                self._arrayH = self.array.T.conj()

            arrOutput = self._arrayH.dot(arrInput)
        else:
            # call fast transform with either cython or python style
            if self._cythonCall:
                # Create output array
                arrOutput = _arrEmpty(
                    2, self.numM, arrInput.shape[1] if ndimInput > 1 else 1,
                    typeInfo[typeOutput].typeNum)

                # Call calculation routine (fused type dispatch must be
                # be done locally in each class, typeInfo gets passed)
                self._backwardC(arrInput, arrOutput, typeInput, typeOutput)
            else:
                arrOutput = self._backward(arrInput)

        if ndimInput == 1:
            # reshape back to single-dimensional vector
            arrOutput = _arrReshape(arrOutput, 1, self.numM, 1, np.NPY_ANYORDER)

        return arrOutput

    ############################################## class reference
    cpdef np.ndarray reference(self):
        '''
        Return an explicit representation of the matrix without using any
        fastmat code. Provides type checks and raises errors if the matrix type
        (self.dtype) cannot hold the reference data.
        '''
        cdef np.ndarray arrRes

        # self._reference() may be overwritten by a child class with either a
        # reference ndarray or a function returning one.

        # test if cython extension-type class method
        callables = [types.FunctionType, types.BuiltinFunctionType,
                     types.MethodType, types.BuiltinMethodType]

        # python2 offers extra unboundMethod distinction
        if sys.version_info < (3, 0):
            callables.append(types.UnboundMethodType)

        # Retrieve this reference array by checking for a ndarray, python or
        # cython-extension-type function
        objSelf = <object> self
        if isinstance(objSelf._reference, np.ndarray):
            arrRes = <np.ndarray> objSelf._reference
        elif isinstance(objSelf._reference, tuple(callables)):
            arrRes = objSelf._reference()
        else:
            try:
                # test if python method
                getattr(objSelf, '__call__')
                arrRes = objSelf._reference()
            except AttributeError:
                arrRes = <np.ndarray> objSelf._reference

        return arrRes

    cpdef np.ndarray _reference(self):
        '''
        Return matrix representation without using a single bit of fastmat code.
        Overwrite this method in child classes to define its reference.
        '''
        return self._array

    def _forwardReferenceInit(self):
        self._forwardReferenceMatrix = self.reference()

    def _forwardReference(self,
                          arrX
                          ):
        '''Calculate the forward transform by non-fastmat means.'''
        if self._forwardReferenceMatrix is None:
            self._forwardReferenceInit()

        return self._forwardReferenceMatrix.dot(arrX)

    ############################################## class inspection, QM
    def _getTest(self):
        from .inspect import TEST, dynFormat
        if self.__class__ == Matrix:
            # Test code for Matrix base class
            return {
                TEST.COMMON: {
                    TEST.NUM_N      : 35,
                    TEST.NUM_M      : TEST.Permutation([30, TEST.NUM_N]),
                    'mType'         : TEST.Permutation(TEST.ALLTYPES),
                    TEST.PARAMALIGN : TEST.Permutation(TEST.ALLALIGNMENTS),
                    'arrM'          : TEST.ArrayGenerator({
                        TEST.DTYPE  : 'mType',
                        TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_M),
                        TEST.ALIGN  : TEST.PARAMALIGN
                    }),
                    TEST.OBJECT     : Matrix,
                    TEST.INITARGS   : (lambda param: [param['arrM']()]),
                    TEST.NAMINGARGS : dynFormat("%s", 'arrM')
                },
                TEST.CLASS: {},
                TEST.TRANSFORMS: {}
            }
        elif isinstance(self, (Hermitian, Conjugate, Transpose)):
            # Test code for the three Transposition classes that are also
            # defined in this submodule. As the Transpositions are directly
            # derived from the Matrix base class we can put the relevant code
            # directly in here and adapt it to work for all three classes.

            # Note that the resulting matrix shape differs for Hermitian and
            # Transpose so we must reflect this when specifying the shape of the
            # to-be-generated data array. As Permutations of a dimension is not
            # uncommon the corresponding fields must be linked in DATASHAPE
            numN = 35
            numM = 30
            swap = isinstance(self, Conjugate)
            return {
                TEST.COMMON: {
                    TEST.NUM_N      : numN,
                    TEST.NUM_M      : TEST.Permutation([numM, TEST.NUM_N]),
                    'mType'         : TEST.Permutation(TEST.ALLTYPES),
                    TEST.PARAMALIGN : TEST.Permutation(TEST.ALLALIGNMENTS),
                    'arrM'          : TEST.ArrayGenerator({
                        TEST.DTYPE  : 'mType',
                        TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_M),
                        TEST.ALIGN  : TEST.PARAMALIGN
                    }),
                    TEST.DATASHAPE  : (TEST.NUM_M if swap else TEST.NUM_N,
                                       TEST.DATACOLS),
                    TEST.DATASHAPE_T: (TEST.NUM_N if swap else TEST.NUM_M,
                                       TEST.DATACOLS),
                    TEST.OBJECT     : self.__class__,
                    TEST.INITARGS   : (lambda param: [Matrix(param['arrM']())]),
                    TEST.NAMINGARGS : dynFormat("%s", 'arrM')
                },
                TEST.CLASS: {},
                TEST.TRANSFORMS: {}
            }
        else:
            # Any other class should go and define its own tests!
            return {}

    def _getBenchmark(self):
        from .inspect import BENCH
        if self.__class__ == Matrix:
            # Benchmark code for Matrix base class
            return {
                BENCH.FORWARD: {
                    BENCH.FUNC_GEN :
                        (lambda c: Matrix(np.zeros((c, c)))),
                },
                BENCH.OVERHEAD: {
                    BENCH.FUNC_GEN  :
                        (lambda c: Matrix(np.zeros((2 ** c, 2 ** c)))),
                    BENCH.FUNC_SIZE : (lambda c: 2 ** c)
                },
                BENCH.DTYPES: {
                    BENCH.FUNC_GEN  :
                        (lambda c, dt: Matrix(np.zeros((2 ** c, 2 ** c),
                                                       dtype=dt)))
                }
            }
        elif isinstance(self, (Hermitian, Conjugate, Transpose)):
            # Benchmark code for the three Transposition classes that are also
            # defined in this submodule. As the Transpositions are directly
            # derived from the Matrix base class we can put the relevant code
            # directly in here and adapt it to work for all three classes.
            # Also note, that we can use the Eye again as the property override
            # of Eye.{T,conj,H} does not grip if instantiated directly through
            # the class constructor
            from .Eye import Eye
            return {
                BENCH.OVERHEAD: {
                    BENCH.FUNC_GEN  :
                        (lambda c: self.__class__(Eye(2 ** c) * 1j)),
                    BENCH.FUNC_SIZE : (lambda c: 2 ** c)
                }
            }
        else:
            # Any other class should go and define its own benchmarks!
            return {}

    def _getDocumentation(self):
        from .inspect import DOC
        return DOC.SUBSECTION(
            r'General Matrix (\texttt{fastmat.Matrix})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
This class serves several purposes. The first one is, that it provides the
functionality to also work with unstructured matrices in conjunction with the
other structured matrices. This is enabled by simply being a wrapper around a
standard NumPy Array \cite{mat_walt2011numpy}.

Second, it serves as the base where all other clases are derived from. This
means it contains properties and methods that are shared across all other \fm{}
classes.""",
                DOC.SNIPPET('# import the package',
                            'import fastmat as fm',
                            '',
                            '# define the parameter',
                            'n = 10',
                            '',
                            '# construct the matrix',
                            'I = fm.Matrix(np.zeros((4,4)))',
                            caption=r"""
This yields a $4 \times 4$ zero matrix which is represented by a \np{} array
and as such does not make use of the structural information anymore."""),
            ),
            DOC.SUBSUBSECTION(
                'Properties', r"""
\begin{itemize}
    \item{Mathematical properties of Matrices}
    \item{Result generated on first request only, then cached}
    \item{Defined in Baseclass, available in all subclasses}
    \item{Behaviour override in child classes possible}
\end{itemize}

These are the properties currently available for all matrices across \fm{}.
\begin{itemize}
\item Transpose (\texttt{Matrix.T})
\item Hermitian Transpose (\texttt{Matrix.H})
\item Conjugate (\texttt{Matrix.conj})
\item Normalized Matrix (\texttt{Matrix.normalized})
\item Gram Matrix (\texttt{Matrix.gram})
\item Largest Eigenvalue (\texttt{Matrix.largestEV})
\end{itemize}"""
            ),
            DOC.SUBSUBSECTION(
                r'Largest Singular Value (\texttt{Matrix.largestSV})', r"""
For a given matrix $\bm A \in \C^{n \times n}$, so $\bm A$ square, we calculate
the absolute value of the largest eigenvalue $\lambda \in \C$. The eigenvalues
obey the equation
\begin{align}
\bm A \cdot \bm v= \lambda \cdot \bm v,
\end{align}
where $\bm v$ is a non-zero vector.

\begin{itemize}
\item \textbf{Input:} matrix $\bm A$, parameter $0 < \varepsilon \ll 1$ as a
stopping criterion
\item \textbf{Output:} largest eigenvalue $\sigma_{\rm max}(\bm A)$
\end{itemize}

\textit{Note:} This algorithm performs well if the two largest eigenvalues are
not very close to each other on a relative scale with respect to their absolute
value. Otherwise it might get trouble converging properly.""",
                DOC.SNIPPET('# import the packages',
                            'import numpy.linalg as npl',
                            'import numpy as np',
                            'import fastmat as fm',
                            '',
                            '# define the matrices',
                            'n = 5',
                            'H = fm.Hadamard(n)',
                            'D = fm.Diag(np.linspace('
                            '        1, 2 ** n, 2 ** n))',
                            '',
                            'K1 = fm.Product(H, D)',
                            'K2 = K1.array',
                            '',
                            '# calculate the eigenvalue',
                            'x1 = K1.largestEV',
                            'x2 = npl.eigvals(K2)',
                            'x2 = np.sort(np.abs(x2))[-1]',
                            '',
                            '# check if the solutions match',
                            'print(x1 - x2)',
                            center=r"""
We define a matrix-matrix product of a Hadamard matrix and a diagonal matrix.
Then we also cast it into a \texttt{numpy}-array and use the integrated EVD. For
demonstration, try to increase $n$ to $>10$ and see what happens.""")
            ),
            DOC.SUBSUBSECTION(
                r'Largest Singular Value (\texttt{Matrix.largestSV})', r"""
For a given matrix $\bm A \in \C^{n \times m}$, we calculate the largest
singular value $\sigma_{\rm max}(\bm A) > 0$, which is the largest entry of the
diagonal matrix $\bm \Sigma \in \C^{n \times m}$ in the decomposition
\begin{align}
\bm A = \bm U \bm \Sigma \bm V^\herm,
\end{align}
where $\bm U$ and $\bm V$ are matrices of the appropriate dimensions. This is
done via the so called power iteration of $\bm A^\herm \cdot \bm A$.

\begin{itemize}
\item \textbf{Input:} matrix $\bm A$, parameter $0 < \varepsilon \ll 1$ as a
stopping criterion
\item \textbf{Output:} largest singular value $\sigma_{\rm max}(\bm A)$
\end{itemize}

\textit{Note:} This algorithm performs well if the two largest singular values
are not very close to each other on a relative scale. Otherwise it might get
trouble converging properly.""",
                DOC.SNIPPET('# import packages',
                            'import numpy.linalg as npl',
                            'import numpy as np',
                            'import fastmat as fm',
                            '',
                            '# define involved matrices',
                            'n = 5',
                            'H = fm.Hadamard(n)',
                            'F = fm.Fourier(2**n)',
                            'K1 = fm.Kron(H, F)',
                            'K2 = K1.array',
                            '',
                            '# calculate the largest SV',
                            '# and a reference solution',
                            'x1 = largestSV(K1.largestSV',
                            'x2 = npl.svd(K2,compute_uv=0)[0]',
                            '',
                            '# check if they match',
                            'print(x1-x2)',
                            caption=r"""
We define a Kronecker product of a Hadamard matrix and a Fourier matrix. Then
we also cast it into a \texttt{numpy}-array and use the integrated SVD. For
demonstration, try to increase $n$ to $>10$ and see what happens.""")),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm M = \bm 0_{2^k \times 2^k}$; so $n = 2^k$ for $k \in \N$""",
                DOC.PLOTFORWARD(),
                DOC.PLOTFORWARDMEMORY(),
                DOC.PLOTOVERHEAD(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            ),
            DOC.BIBLIO(
                mat_walt2011numpy=DOC.BIBITEM(
                    r"""
St\'efan van der Walt, S. Chris Colbert and Ga\"el Varoquaux""",
                    r"""
The NumPy Array: A Structure for Efficient Numerical Computation""",
                    r"""
Computing in Science and Engineering, Volume 13, 2011.""")
            )
        ) if self.__class__ == Matrix else ""

################################################################################
################################################## Hermitian transpose class
cdef class Hermitian(Matrix):

    ############################################## class methods
    def __init__(self, Matrix matrix):
        if not isinstance(matrix, Matrix):
            raise TypeError("Hermitian: Not a fastmat Matrix")

        self._nested = matrix
        self._content = (matrix, )
        self._initProperties(matrix.shape[1], matrix.shape[0], matrix.dtype,
                             cythonCall=matrix._cythonCall,
                             widenInputDatatype=matrix._widenInputDatatype,
                             forceInputAlignment=matrix._forceInputAlignment)

    def __repr__(self):
        '''
        Return a string representation of this class instance.
        The __repr__() method of the nested transformation gets extended by an
        info about the applied transposition.
        '''
        return "<%s.H>" %(self._nested.__repr__())

    ############################################## class property override
    cpdef np.ndarray _getArray(self):
        return self._nested._getArray().T.conj()

    cpdef np.ndarray _getCol(self, intsize idx):
        return _conjugate(self._nested._getRow(idx))

    cpdef np.ndarray _getRow(self, intsize idx):
        return _conjugate(self._nested._getCol(idx))

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return np.conjugate(self._nested._getItem(idxM, idxN))

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return self._nested.largestEV

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return self._nested.largestSV

    cpdef Matrix _getT(self):
        return getConjugate(self._nested)

    cpdef Matrix _getH(self):
        return self._nested

    cpdef Matrix _getConj(self):
        return Transpose(self._nested)

    ############################################## class performance estimation
    cpdef tuple _getComplexity(self):
        return (self.numM + self.numN, self.numN + self.numM)

    ############################################## class forward / backward
    cpdef _forwardC(self, np.ndarray arrX, np.ndarray arrRes,
                    ftype typeX, ftype typeRes):
        '''Calculate the forward transform of this matrix, cython-style.'''
        self._nested._backwardC(arrX, arrRes, typeX, typeRes)

    cpdef _backwardC(self, np.ndarray arrX, np.ndarray arrRes,
                     ftype typeX, ftype typeRes):
        '''Calculate the backward transform of this matrix, cython-style.'''
        self._nested._forwardC(arrX, arrRes, typeX, typeRes)

    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''
        return self._nested._backward(arrX)

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''
        return self._nested._forward(arrX)

    ########################################## references: test / benchmark
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using any
        fastmat code.
        '''
        return self._nested.reference().T.conj()


################################################################################
################################################## Conjugate factory
cdef inline Matrix getConjugate(Matrix matrix):
    '''
    Return the conjugate of matrix if matrix has a complex data type. Otherwise
    return matrix. Acts as factory for the Conjugate metaclass.
    '''
    return (Conjugate(matrix) if matrix._info.dtype[0].isComplex else matrix)


################################################################################
################################################## Conjugate class
cdef class Conjugate(Matrix):

    ############################################## class methods
    def __init__(self, Matrix matrix):
        if not isinstance(matrix, Matrix):
            raise TypeError("Conjugate: Not a fastmat Matrix")

        self._nested = matrix
        self._content = (matrix, )
        self._initProperties(matrix.shape[0], matrix.shape[1], matrix.dtype,
                             cythonCall=matrix._cythonCall,
                             widenInputDatatype=matrix._widenInputDatatype,
                             forceInputAlignment=matrix._forceInputAlignment)

    def __repr__(self):
        '''
        Return a string representation of this class instance.
        The __repr__() method of the nested transformation gets extended by an
        info about the applied transposition.
        '''
        return "<conj(%s)>" %(self._nested.__repr__())

    ############################################## class property override
    cpdef np.ndarray _getArray(self):
        return self._nested._getArray().conj()

    cpdef np.ndarray _getCol(self, intsize idx):
        return _conjugate(self._nested._getCol(idx))

    cpdef np.ndarray _getRow(self, intsize idx):
        return _conjugate(self._nested._getRow(idx))

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return np.conjugate(self._nested._getItem(idxN, idxM))

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return self._nested.largestEV

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return self._nested.largestSV

    cpdef Matrix _getT(self):
        return Hermitian(self._nested)

    cpdef Matrix _getH(self):
        return Transpose(self._nested)

    cpdef Matrix _getConj(self):
        return self._nested

    ############################################## class performance estimation
    cpdef tuple _getComplexity(self):
        cdef float complexity = self.numN + self.numM
        return (complexity, complexity)

    ############################################## class forward / backward
    cpdef _forwardC(self, np.ndarray arrX, np.ndarray arrRes,
                    ftype typeX, ftype typeRes):
        '''Calculate the forward transform of this matrix, cython-style.'''
        cdef np.ndarray arrInput = _conjugate(arrX)
        self._nested.forwardC(arrInput, arrRes, typeX, typeRes)
        _conjugateInplace(arrRes)

    cpdef _backwardC(self, np.ndarray arrX, np.ndarray arrRes,
                     ftype typeX, ftype typeRes):
        '''Calculate the backward transform of this matrix, cython-style.'''
        cdef np.ndarray arrInput = _conjugate(arrX)
        self._nested.backwardC(arrInput, arrRes, typeX, typeRes)
        _conjugateInplace(arrRes)

    cpdef np.ndarray _forward(self, np.ndarray arrX):
        '''Calculate the forward transform of this matrix'''
        cdef np.ndarray arrRes = self._nested._forward(_conjugate(arrX))
        _conjugateInplace(arrRes)
        return arrRes

    cpdef np.ndarray _backward(self, np.ndarray arrX):
        '''Calculate the backward transform of this matrix'''
        cdef np.ndarray arrRes = self._nested._backward(_conjugate(arrX))
        _conjugateInplace(arrRes)
        return arrRes

    ########################################## references: test / benchmark
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using any
        fastmat code.
        '''
        return self._nested._reference().conj()


################################################################################
################################################## Transpose class
cdef class Transpose(Hermitian):
    ############################################## class methods
    def __init__(self, Matrix matrix):
        if not isinstance(matrix, Matrix):
            raise TypeError("Transpose: Not a fastmat Matrix")

        # NOTE: Transpose is implemented as a Hermitian of a Conjugated Matrix.
        # Thus the _nested field points to the Conjugate of the actual base
        # matrix. For working on the true base matrix, _nestedConj must be
        # accessed instead of _nested like in Hermitian or Conjugated.
        self._nestedConj = matrix
        matrix = getConjugate(matrix)

        self._nested = matrix
        self._content = (matrix, )
        self._initProperties(matrix.shape[1], matrix.shape[0], matrix.dtype,
                             cythonCall=matrix._cythonCall,
                             widenInputDatatype=matrix._widenInputDatatype,
                             forceInputAlignment=matrix._forceInputAlignment)

    def __repr__(self):
        '''
        Return a string representation of this class instance.
        The __repr__() method of the nested transformation gets extended by an
        info about the applied transposition.
        '''
        return "<%s.T>" %(self._nestedConj.__repr__())

    ############################################## class property override
    cpdef np.ndarray _getArray(self):
        return self._nestedConj._getArray().T

    cpdef np.ndarray _getCol(self, intsize idx):
        return self._nestedConj._getRow(idx)

    cpdef np.ndarray _getRow(self, intsize idx):
        return self._nestedConj._getCol(idx)

    cpdef object _getItem(self, intsize idxN, intsize idxM):
        return self._nestedConj._getItem(idxM, idxN)

    cpdef object _getLargestEV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return self._nestedConj.largestEV

    cpdef object _getLargestSV(self, intsize maxSteps,
                               float relEps, float eps, bint alwaysReturn):
        return self._nestedConj.largestSV

    cpdef Matrix _getT(self):
        return self._nestedConj

    cpdef Matrix _getH(self):
        return getConjugate(self._nestedConj)

    cpdef Matrix _getConj(self):
        return Hermitian(self._nestedConj)

    ########################################## references: test / benchmark
    cpdef np.ndarray _reference(self):
        '''
        Return an explicit representation of the matrix without using any
        fastmat code.
        '''
        return self._nestedConj._reference().T



# -*- coding: utf8 -*-
'''
  demo/compOmpIsta.py
 -------------------------------------------------- part of the fastmat demos


  Author      : wcw
  Introduced  :
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''

import sys                                        # system calls
import time                                       # measuring time

import numpy                                      # math library
import scipy.sparse                               # math library


################################################## import modules
try:
    import matplotlib.pyplot as plt               # plotting
    import matplotlib.gridspec as gridspec        # grid specification
except ImportError:
    print("matplotlib not found. Please consider installing it to proceed.")
    sys.exit(0)

try:
    import fastmat
except ImportError:
    sys.path.append('..')
    import fastmat


################################################## helper functions
def printTime(note, call, *args, **kwargs):
    '''Measure runtime of call and print it to stdout.'''
    sys.stdout.write(" * %s ... " % (note))
    sys.stdout.flush()

    # measure runtime of call, print it and return result
    t = time.time()
    result = call(*args, **kwargs)
    print("%.2e s" % (time.time() - t))
    return result


def plotComplex(plot, vec):
    '''Plot real and imaginary part of vector concatenated along x'''
    plot.axes.get_xaxis().set_visible(False)
    plot.imshow(
        numpy.concatenate([vec.real, vec.imag], axis=1),
        interpolation='none',
        aspect='auto'
    )

################################################## script


print("fastmat demo: Sparse-Signal-Recovery Demo using OMP and ISTA")
print("------------------------------------------------------------")

# define some constants
N = 100        # number of measurements (height of dictionary)
M = 500        # width of dictionary
K = 30        # sparsity (non-zero components in support)

# define baseline random support (1.0 forces dtype to be float)
s = scipy.sparse.rand(M, 1, 1.0 * K / M).todense()
x0 = s - 1j * s

# define some dictionary
# (dictionary matrix holds first N entries of M-Fourier matrix)
# option to choose a complex one or a non-complex one
matG = fastmat.Matrix(
        numpy.random.normal(0.0, numpy.sqrt(N), (N, M)) + \
        1j * fastmat.Matrix(numpy.random.normal(0.0, numpy.sqrt(N), (N, M)))
    )

# option 1
mat = fastmat.Product(
    matG,
    fastmat.Fourier(M)
)

# option 2
#mat = matG

# generate measurements from baseline support
b = mat * x0

# run OMP and ISTA
# result = fastmat.OMP(mat, b, K)
xOMP = printTime("running OMP",
                 fastmat.algs.OMP, mat, b, K)
# result = fastmat.ISTA(mat, b, numLambda=1e4)
xISTA = printTime("running ISTA",
                  fastmat.algs.ISTA, mat, b,
                  numLambda = 1e6, numMaxSteps = 1000)

# plot:
fig = plt.figure()
grid = gridspec.GridSpec(
    2, 3,
    height_ratios=[1, 7],
    width_ratios=[M / (2 * N), M / (2 * N), 1])

# plot dictionary. as it is a partial Fourier, only plot element phases
plot = fig.add_subplot(grid[0, :-1], title='Dictionary matrix').imshow(
    numpy.angle(mat.array),
    interpolation='none', aspect='auto')

# plot measurements next to matrix
plotComplex(fig.add_subplot(grid[0, -1], title='Measurements'), b)

# plot support vectors (baseline left, OMP center, ISTA right)
plotComplex(fig.add_subplot(grid[1, 0], title='OMP (real|imag)'), xOMP)
plotComplex(fig.add_subplot(grid[1, 1], title='ISTA (real|imag)'), xISTA)
plotComplex(fig.add_subplot(grid[1, 2], title='baseline'), x0)

# set title of whole figure
fig.suptitle(
    "Sparse Signal Recovery Demo with OMP and ISTA using fastmat",
    fontweight='bold')

# arrange tightly and plot
grid.tight_layout(fig)
plt.show()



# -*- coding: utf-8 -*-
'''
  util/bee.py
 -------------------------------------------------- part of the fastmat package

  Utility for shell interaction with class code from python packages.

  Usecases:
    - find all modules in a (sub-)package, which contain certain functions
        --check-only --function

    - provide a string interface to module-level functions enabling shell
      piping to python package internals
        --argument --function

    - flexible options for character encoding
        --encode utf-8

  Example:  retrieve LaTeX documentation of class Circulant, execute the
            following one-liner from the doc/ directory of fastmat:

            python doc_extract.py --path .. --package fastmat --name Circulant
            --function docLaTeX --encode utf-8


  Author      : wcw
  Introduced  : 2017-01-03
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import sys
import inspect
import os
import argparse
from pprint import pprint
import time

import numpy as np


def importMatplotlib():
    '''Conditional import of matplotlib'''
    try:
        from matplotlib import pyplot as plt
    except ImportError:
        print("matplotlib not found. Please consider installing it to proceed.")
        sys.exit(0)

    return plt


colors = ['#003366', '#FF6600', '#CC0000', '#FCE1E1']

################################################## import fastmat
try:
    import fastmat
except ImportError:
    sys.path.append('..')
    sys.path.append('.')
    import fastmat

from fastmat.inspect import Test, Benchmark, Documentation, TEST, BENCH, DOC
from fastmat.inspect.common import AccessDict, \
    fmtBold, fmtGreen, fmtYellow, fmtRed

classBaseContainers = (fastmat.Matrix, fastmat.Algorithm)

################################################## package constant definition
packageName = 'fastmat'
toolTitle = 'FastMat command line package interface tool (working Bee)'

EXT_BENCH_RESULT = 'csv'
EXT_TEST_RESULT = 'mat'

################################################################################
##################################################  subcommand argparser


def getObjects(
    obj,
    aSelector=(lambda obj, name: (name[0] != '_'))
):
    '''
    Return list of objects in obj whose properties satisfy aSelector.

    aClass may be a class type or a module. aSelector is a lambda function,
    filtering private objects whose names start with an underscore ('_').
    '''
    return sorted([item for item in list(obj.__dict__.keys())
                   if aSelector(obj, item)])


def unpackExtraArgs(extraArgs):
    result = {}
    if extraArgs is not None:
        for p in extraArgs:
            for pp in list(p.split(',')):
                tokens = pp.split('=')
                if len(tokens) >= 2:
                    string = "=".join(tokens[1:])
                    try:
                        val = int(string)
                    except ValueError:
                        try:
                            val = float(string)
                        except ValueError:
                            val = string
                    result[tokens[0]] = val
    return result


class CommandArgParser(object):
    _commandType = '<command>'

    def __init__(self, *arguments, **options):
        '''Detect commands and dispatch to specific parsers for each.'''

        # store passed (and already processed portion of command line)
        prevArgs = options.get('prevArgs', '')
        self._prevArgs = prevArgs

        # configure parser
        validCmds = '|'.join(getObjects(self.__class__))
        self.cmdParser = argparse.ArgumentParser(
            usage="%s %s [arguments]" % (prevArgs, self._commandType))

        # first-pass argument.parse_args defaults to [1:], selecting command
        self.cmdParser.add_argument(self._commandType, help=validCmds)
        argument = arguments[0] if len(arguments) > 0 else ''
        self._args = self.cmdParser.parse_args((argument,))

        # create argument parser for actual subcommands
        self.argParser = argparse.ArgumentParser(
            usage="%s %s [arguments]" % (prevArgs, argument))

        # try to call method with this name
        try:
            # dispatch all other arguments to subcommand-stage
            dispatch = getattr(self, argument)
        except AttributeError:
            print(os.linesep + toolTitle)
            self.cmdParser.print_help()
            exit(1)

        # for the next stage subcommands handling, pass current arguments too.
        if inspect.isclass(dispatch):
            dispatch(*(arguments[1:]), prevArgs=' '.join([prevArgs, argument]))
        else:
            dispatch(*(arguments[1:]))


################################################################################
################################################## command line argument parser
class Bee(CommandArgParser):

    ############################################## command: list
    class list(CommandArgParser):
        _commandType = '<option>'

        def _parseArgs(self, *arguments):
            '''Process further cmd-line arguments, store them in self.args.'''
            self.argParser.add_argument(
                '-e', '--extended',
                action='store_true',
                help='Show extended information'
            )
            self.args = self.argParser.parse_args(arguments)

        def _printDataset(self, dataset, newline='\n', separator=' '):
            '''List contents of dataset according to modifiers in self.args.'''
            if self.args.extended:
                pprint(dataset)
            else:
                # extract all key names of dataset and print the list
                print(separator.join(
                    sorted([loc for loc in dataset.keys()])) + newline)

        def _parseAndPrint(self, dataset, *arguments):
            '''Combine self._parseArgs and self._printDataset.'''
            self._parseArgs(*arguments)
            self._printDataset(dataset)

        def classes(self, *arguments):
            '''Return list of available classes in package.'''
            self._parseAndPrint(packageClasses, *arguments)

        def algs(self, *arguments):
            '''Return list of available algorithms in package.'''
            self._parseAndPrint(packageAlgs, *arguments)

        def index(self, *arguments):
            '''Return list of available units in package.'''
            self._parseAndPrint(packageIndexTree, *arguments)

        def makedump(self, *arguments):
            '''Dump all information for makefile processing.'''
            self._parseArgs()

            printSetup = {'newline': ';', 'separator': ':'}
            for dataset in [packageClasses, packageAlgs]:
                self._printDataset(dataset, **printSetup)

    ############################################## selection helper function
    def _select(self, nameType, arguments, lstElements):
        self.argParser.add_argument(
            '-s', '--select',
            nargs='+',
            default=[],
            help=("Allows customizing the set of %(name)s to be run by " +
                  "combining filtering with picking. Append an arbitrary " +
                  "number of selectors in the format 'unit.%(name)s'. " +
                  "Selectors specifying only one category ('unit.' or " +
                  "'.%(name)s') introduce mask filter capabilities, entries. " +
                  "If any mask filters are specified, all targets matching " +
                  "both the combined unit and the combined %(name)s masks " +
                  "will be added to the job. Any explicitly picked targets " +
                  "('unit.%(name)s') will be added to the job regardless of " +
                  "also specified mask filters. If no mask filters are " +
                  "specified, but some targets are picked explicitly, only " +
                  "the picked targets will be %(name)sed. If neither mask " +
                  "filters are specified nor targets are pciked, all " +
                  "available targets will be %(name)sed. Arguments with no " +
                  "separators ('.') will be ignored completely.") % {
                'name': nameType}
        )
        self.argParser.add_argument(
            '-o', '--options',
            nargs='+',
            default=[],
            help=("You may specify additional parameters for the %(name)ss " +
                  "by putting extra <NAME>=<VALUE> either with this option " +
                  "or where they do not interfere with other options. You " +
                  "may separate multiple parameters either with whitespaces " +
                  "or commas.") % {'name': nameType}
        )

        # parse arguments, all extra stuff goes to 'extraParam'
        self.args, self.extraArgs = self.argParser.parse_known_args(arguments)

        # add extraParams added with the -o option
        self.extraArgs = tuple(list(self.extraArgs) + self.args.options)

        # compile job list (--select argument), default: all
        # prepare a dictionary with {element:[]} structure to keep track of all
        # selected targets for each element.
        if len(self.args.select) == 0:
            # This defaults to 'select all targets of all elements'
            return {element: [] for element in lstElements}
        else:
            # default to no selected elements(every element selects one element
            # representing an illegal target -> nothing will be tested unless
            # another element is added
            dummyElement = ''
            selected = {element: set([dummyElement]) for element in lstElements}

            # prepare lists to collect parameters
            selectElement = set([])

            # classify options in lists
            for name in self.args.select:
                tokens = name.split('.')
                if len(tokens) < 2:
                    continue

                element, target = '.'.join(tokens[:-1]), tokens[-1]
                if tokens[0] == '':
                    # begins with '.' -> ".target" selection mask
                    # if specific targets were selected, add them to all
                    # elements resulting in full selection of target
                    for lstTargets in selected.values():
                        lstTargets.add(target)
                elif tokens[-1] == '':
                    # ends on '.' -> "element." selection mask
                    # collect these elements for now
                    selectElement.add(element)
                else:
                    # element.target specific selection mask
                    # only select this specific element:target pair
                    if element in lstElements:
                        selected[element].add(tokens[-1])

            # to select some elements completely the selection list of that
            # particular element only needs to become empty
            for element in selectElement:
                if element in lstElements:
                    selected[element].clear()

            # now remove all elements containing only the dummy element (these
            # were actually not selected at all)
            selected = {
                element: lstTargets
                for element, lstTargets in selected.items()
                if len(lstTargets) != 1 or dummyElement not in lstTargets}

            # as cleanup stage remove the empty-blocking '' elements from all
            # sets which contain one other parameter as well (> two elements)
            # also do convert bayk to list
            for element, lstElements in selected.items():
                if len(lstElements) > 1:
                    lstElements.discard(dummyElement)

                selected[element] = sorted(list(selected[element]))

            # return the selector as list
            return selected

    ############################################## command: test
    def test(self, *arguments):
        '''Run unit tests in package.'''
        # parse arguments and get selection from list
        self.argParser.add_argument(
            '-w', '--write-failed',
            action='store_true',
            help="If specified, stores failed test results to a .mat file " +
            "for each failed test instance, including parameters, input and " +
            "evaluation output"
        )
        self.argParser.add_argument(
            '-i', '--interact',
            action='store_true',
            help="Start interactive session for investigating the identified " +
            "problems"
        )
        self.argParser.add_argument(
            '-v', '--verbose',
            action='store_true',
            help="Increase verbosity of output."
        )
        self.argParser.add_argument(
            '-f', '--full',
            action='store_true',
            help="Produce full output."
        )
        self.argParser.add_argument(
            '-c', '--calibration',
            type=str,
            default='',
            help="Filename to read calibration data from. Loads calibration " +
            "data for fastmat classes which is then used by the package."
        )

        # stop time
        timeTotal = time.time()

        # determine the selection of classes-under-test and their tergets
        selection = self._select(TEST.TEST, arguments, packageIndex.keys())

        # sanity check
        if len(selection) < 1:
            self.argParser.exit(0, "test: job list is empty.\n")

        # generate test worker classes
        options = unpackExtraArgs(self.extraArgs)
        tests = {name: Test(packageIndex[name], extraOptions=options)
                 for name in selection.keys()}

        # if workers have no targets, discard them (evaluated during __init__)
        tests = {name: worker
                 for name, worker in tests.items()
                 if len(worker.options.keys()) > 0}

        # run tests
        cntTests, cntProblems, cntIrregularities = [0], [0], [0]
        for nameTest, test in sorted(tests.items()):
            # measure timing as lap timer in callback
            timeLap = [time.time()]

            def cbResult(name, targetResult):
                # runtime of last test
                timeSingle = time.time() - timeLap[0]

                # statistics and output
                test.findProblems(name, targetResult)
                numProb = len(test.problems[name])
                numIrr = len(test.irregularities[name])
                numTests = sum(sum(len(variant.values())
                                   for variant in test.values())
                               for test in targetResult.values())
                cntProblems[0] += numProb
                cntIrregularities[0] += numIrr
                cntTests[0] += numTests

                print((fmtBold(">> %s:") + " ran %d tests in %.2f seconds") %(
                    '.'.join([nameTest, name]), numTests, timeSingle
                ) + " (%s%s)" %(
                    (fmtGreen if numProb == 0 else fmtRed)
                    ("%d problems" % (numProb)),
                    ("" if numIrr == 0
                     else fmtYellow(
                         ", %d irregularities (ignored)" %(numIrr)))
                ))
                # refresh timer for next measurement
                timeLap[0] = time.time()

            # set callbacks, verbosity and start testing all selected targets
            test.verbosity = self.args.verbose, self.args.full
            test.cbResult = cbResult
            test.run(*(selection[nameTest]))

        # forget about all tests with no targets executed (the selection list
        # did not match the targets offered by the test) to exclude it from
        # the unit result reporting following
        tests = {name: test
                 for name, test in tests.items()
                 if len(test.results.keys()) > 0}

        # analyze dependencies of units involved in problems
        testClasses = {name: packageIndex[name]
                       for name in tests.keys()}
        # STAGE 1: detect structural dependencies (crawl inheritance model)
        # a dependency is a class of which the respective class is a subclass of
        # limit search to classes derived from one of the class containers
        testDependencies = {
            name: set(inspect.getmro(classType))
            for name, classType in testClasses.items()}

        # STAGE 2: add dependencies introduced in test case instances
        def crawlContent(item, targetSet):
            for nestedItem in item:
                targetSet.add(nestedItem.__class__)
                crawlContent(nestedItem, targetSet)

        for name, testWorker in tests.items():
            targetSet = testDependencies[name]
            for target in testWorker.results.values():
                for nameTest, test in target.items():
                    if len(test) > 0:
                        aVariant = list(test.values())[0]
                        if len(aVariant) > 0:
                            aQuery = list(aVariant.values())[0]
                            crawlContent(aQuery.get(TEST.INSTANCE, ()),
                                         targetSet)

        # filter out the containers themselves as they are obviousely the
        # "mothers of it all" and as each class is also a subclass to itself,
        # skip it as well to improve readability of output
        testDependencies = {
            name: [item
                   for item in dependencies
                   if (issubclass(item, classBaseContainers) and
                       item not in classBaseContainers and
                       item != testClasses[name])]
            for name, dependencies in testDependencies.items()}

        # split units into two main lists (unit has problems / has no problems)
        problemUnits = [name
                        for name, test in sorted(tests.items())
                        if sum(len(problems)
                               for problems in test.problems.values()) > 0]
        fineUnits = [name
                     for name in sorted(tests.keys())
                     if name not in problemUnits]

        # now divide the `has problems` list further based on dependencies
        dependentProblems = [name
                             for name in problemUnits
                             if len(testDependencies[name]) > 0]
        independentProblems = [name
                               for name in problemUnits
                               if name not in dependentProblems]

        # stop time, print results and show summary of dependencies
        timeTotal = time.time() - timeTotal
        cntTests = cntTests[0]
        cntProblems, cntIrregularities = (cntProblems[0], cntIrregularities[0])

        print("\nResults:")
        print(("   > found %d problem(s) in %d classes%s " +
               "(ran %d tests in %.2f seconds)") %(
            cntProblems, len(problemUnits),
            ("" if cntIrregularities == 0
             else " (%d issues ignored)" %(cntIrregularities)),
            cntTests, timeTotal))

        if len(fineUnits) > 0:
            print("   > Classes with no problems:")
            print("%s[%s]\n" % (
                " " * 8, ", ".join([fmtGreen(unit)
                                    for unit in fineUnits])))

        if len(independentProblems) > 0:
            print("   > Classes with independent problems (start fixing here):")
            print("%s[%s]\n" % (
                " " * 8, ", ".join([fmtRed(unit)
                                    for unit in sorted(independentProblems)])))

        if len(dependentProblems) > 0:
            nameLength = max(len(name) for name in dependentProblems)
            print("   > Classes with problems that depend on other classes")

            def fmtDummy(string):
                return string

            for name in dependentProblems:
                depList = [((fmtYellow if dep in problemUnits else fmtGreen)
                            if dep in tests.keys() else fmtDummy)(dep.__name__)
                           for dep in testDependencies[name]]
                print("%s%*s: [%s]" % (
                    " " * 8, nameLength, fmtRed(name),
                    ", ".join(sorted(depList))))
            print("\n")

        # make test dictionary easy accessible
        tests = AccessDict(tests)

        # interactive session, anyone?
        if self.args.interact and cntProblems > 0:
            # register signaling for graceful exit
            import atexit

            def quitGracefully():
                print("Leaving interactive testing session.")

            atexit.register(quitGracefully)

            # scope dictionary
            scope = {}

            # set numpy print options
            np.set_printoptions(linewidth=200, threshold=10, edgeitems=3,
                                formatter={'float_kind'     : '{: .3}'.format,
                                           'complex_kind'   : '{: .3e}'.format})

            # scope dictionary
            scope.update(globals())
            scope.update(locals())

            # start interactive session
            import code
            code.interact(r"  > Entering interactive testing session.",
                          None, scope)

        # if problems occurred, exit with non-zero errorcode
        if cntProblems > 0:
            sys.exit(1)

    ############################################## command: benchmark
    def benchmark(self, *arguments):
        '''Run performance evaluation of units in package.'''

        # parse arguments and get selection from list
        self.argParser.add_argument(
            '-p', '--path-results',
            type=str,
            default='.',
            help="Save the benchmark results to the path specified."
        )
        self.argParser.add_argument(
            '-c', '--calibration',
            type=str,
            default='',
            help="Filename to read calibration data from. Loads calibration " +
            "data for fastmat classes which is then used by the package."
        )
        selection = self._select(
            BENCH.BENCHMARK, arguments, packageIndex.keys())

        # load calibration data
        if len(self.args.calibration) > 0:
            fastmat.core.loadCalibration(self.args.calibration)
            print("  > Loaded calibration data from %s." %(
                self.args.calibration))

        # sanity check
        if len(selection) < 1:
            self.argParser.error("benchmark: job list is empty.")

        # generate benchmark worker classes
        benchmarks = {
            name: Benchmark(packageIndex[name],
                            extraOptions=unpackExtraArgs(self.extraArgs))
            for name in selection.keys()}

        # run benchmarks
        for nameBench, bench in sorted(benchmarks.items()):
            def cbResult(name, targetResult):
                # output, anyone?
                if len(self.args.path_results) > 0:
                    print("   > saved result '%s' to '%s'" % (
                        name,
                        bench.saveResult(name, outPath=self.args.path_results)))

                print("")

            # start tests of all targets of this element
            bench.cbResult = cbResult
            bench.run(*(selection[nameBench]))

    ############################################## command: documentation
    def documentation(self, *arguments):
        '''Extract built-in documentation from units in package.'''
        self.argParser.add_argument(
            'units',
            nargs='*',
            default=[],
            help="Specify the documentation sections to include in the TeX " +
            "output. If this list is empty, all sections in the package will " +
            "be considered. Any key=value pairs will be forwarded to " +
            "benchmark operations"
        )
        self.argParser.add_argument(
            '-p', '--path-results',
            type=str,
            default='.',
            help='Path to the benchmark results.'
        )
        self.argParser.add_argument(
            '-o', '--output',
            type=str,
            default='',
            help="Filename to write the output to. If not file is specified " +
            "the output will be directed to STDOUT."
        )
        self.argParser.add_argument(
            '-c', '--calibration',
            type=str,
            default='',
            help="Filename to read calibration data from. Loads calibration " +
            "data for fastmat classes which is then used by the package."
        )

        # parse arguments, all extra stuff goes to 'extraParam'
        self.args, extraArgs = self.argParser.parse_known_args(arguments)
        self.args.units.extend(extraArgs)

        # load calibration data
        if len(self.args.calibration) > 0:
            fastmat.core.loadCalibration(self.args.calibration)
            print("  > Loaded calibration data from %s." %(
                self.args.calibration))

        # compile set of extra options for optional benchmarking
        benchmarkOptions = unpackExtraArgs([name
                                            for name in self.args.units
                                            if '=' in name])
        options = {'benchmarkOptions'   : benchmarkOptions,
                   DOC.OUTPATH          : self.args.path_results}

        selection = [name for name in self.args.units if '=' not in name]

        # if selection is empty, take all sections of package
        if len(selection) == 0:
            selection = (sorted(packageClasses.keys()) +
                         sorted(packageAlgs.keys()))
        else:
            selection = [name for name in selection if name in packageIndex]

        if len(selection) < 1:
            self.argParser.error("documentation: job list is empty.")

        output = ["% THIS OUTPUT WAS GENERATED AUTOMATICALLY AND " +
                  "MAY BE OVERWRITTEN ANYTIME"]

        # add machine info ahead of documentation
        import platform
        machineInfo = (
            " ".join([" \\verb|%s|" %(token)
                      for token in platform.platform().split("-")]).strip("`"),
            " ".join([" \\verb|%s|" %(token)
                      for token in " ".join(
                          platform.uname()).split(" ")]).strip("'"),
            platform.processor(),
            '---'
        )
        output.append(DOC.SUBSECTION('General information', r"""
This section was automatically generated on a system with the following
specifications

\vspace{5mm}\begin{centering}
  \begin{tabular}[t]{ m{.45\columnwidth} | m{.45\columnwidth} }
    \textbf{\large System} & \textbf{\large Kernel} \\
      \begin{flushleft}%s\end{flushleft} & \begin{flushleft}%s\end{flushleft} \\
      \textbf{\large Processor} & \textbf{\large Memory} \\
      \verb|%s| & \verb|%s|
  \end{tabular}
\end{centering}""" % machineInfo))

        # get documentation of selected units
        for name in selection:
            # add class name to tex output and make output open on a new column
            output.extend(['', '%' * 80,
                           "%% Documentation of unit '%s'" %(name),
                           '%' * 80, '',
                           r'\vfill\null\columnbreak'])

            # output class documentation
            output.append(Documentation(packageIndex[name], **options))

        # finally, print output
        strOutput = os.linesep.join([str(item)
                                     if isinstance(item, DOC.__DOC__) else item
                                     for item in output])
        if len(self.args.output) == 0:
            print(strOutput)
        else:
            with open(self.args.output, "w") as f:
                f.write(strOutput)

            print(" >> output written to '%s'" %(self.args.output))

    ############################################## command: calibrate
    def calibrate(self, *arguments):
        '''Extract built-in documentation from units in package.'''
        self.argParser.add_argument(
            'classes',
            nargs='*',
            default=[],
            help="If an output file is specified, select the classes to " +
            "perform a calibration on. If as sole item the string `all` is " +
            "given, all classes in fastmat will be selected."
        )
        self.argParser.add_argument(
            '-i', '--input',
            type=str,
            default='',
            help="Filename to read calibration data from. May be used to " +
            "load existing calibration data for plotting."
        )
        self.argParser.add_argument(
            '-o', '--output',
            type=str,
            default='',
            help="Perform a calibration of all selected classes. Write the " +
            "resulting calibration data to OUTPUT"
        )
        self.argParser.add_argument(
            '-p', '--plot',
            action='store_true',
            help="Plot the calibration data for all selected classes. If no " +
            "new calibration was performed a fresh set of overhead " +
            "benchmarks will be generated and used for plotting against."
        )
        self.argParser.add_argument(
            '-v', '--verbose',
            action='store_true',
            help="Be verbose about the calibration."
        )

        # parse arguments
        self.args = self.argParser.parse_args(arguments)

        # determine a list of selected classes: interpret codewords
        if len(self.args.classes) == 1 and self.args.classes[0] == 'all':
            classes = fastmat.classes
        else:
            classNames = {item.__name__: item for item in fastmat.classes}
            classes = [classNames[name]
                       for name in self.args.classes
                       if name in classNames]

        # explore dependencies of class benchmarks in two stages:
        # STAGE I: determine class inheritance relations (code dependencies)
        dependencies = {
            item: set([child for child in inspect.getmro(item)
                       if (issubclass(child, fastmat.Matrix) and
                           not (child == item))])
            for item in fastmat.classes}

        # STAGE II: add nested matrix classes in benchmark instances
        #           (data dependencies)
        def nodesOfTree(tree):
            nodes = set()
            if tree is not None:
                for child in tree:
                    nodes.add(child)
                    nodes.update(nodesOfTree(child))

            return nodes

        instances = [
            Benchmark(item).options[BENCH.OVERHEAD][BENCH.FUNC_GEN](2)
            for item in fastmat.classes]

        for inst in instances:
            deps = dependencies.setdefault(inst.__class__, set())
            deps.update(node.__class__ for node in nodesOfTree(inst))

        # now compile a new selection list, which also includes dependent but
        # currently not yet calibrated classes
        # traverse the list in order of dependency count to put baseclasses
        # before deeply-nested metalasses. sort alphabetically if count matches
        selected = []
        for item in sorted(classes,
                           key=lambda x: (len(dependencies[x]), x.__name__)):
            # the class in `item` was selected for calibration. Make
            # sure all dependent classes are calibrated as well
            def updateSelected(node):
                # recurse over all subnodes of node
                for dep in dependencies[node]:
                    if dep not in selected:
                        updateSelected(dep)

                # finally, add the node that just got crawled it not
                # already done before
                if (node not in selected and
                        node not in fastmat.core.calData):
                    selected.append(node)

            # add all the requested classes to the job list. Make sure
            # the order is such that no class is added before all of its
            # dependencies are fulfilled by the list contents so far
            updateSelected(item)

        self.args.classes = selected

        # load calibration data from file if specified
        if len(self.args.input) > 0:
            fastmat.core.loadCalibration(self.args.input)
            print("  > Loaded calibration data from %s." %(self.args.input))

        # generate calibration data for the classes specified
        benchmarks = {}
        # first, figure out in which order to perform the calibrations
        # also

        # finally run the benchmarks
        for item in self.args.classes:
            print("  > Running calibration of class '%s'" %(item.__name__))
            cal, bench = fastmat.core.calibrateClass(item,
                                                     verbose=self.args.verbose)
            benchmarks[item] = bench

        # write calibration data to file if an output file is specified
        if len(self.args.output) > 0:
            fastmat.core.saveCalibration(self.args.output)
            print("  > Written calibration data to %s." %(self.args.output))

        # plot the current set of calibration data and perform benchmarks if
        # we have no data for some of the classes
        if self.args.plot:
            missing = [item for item in fastmat.core.calData
                       if item not in benchmarks]
            if len(missing) > 0:
                print("  > Generating performance measurements for class plots")
                for item in missing:
                    print("     - %s" %(item.__name__))
                    benchmarks[item] = fastmat.core.calibrateClass(
                        item, verbose=self.args.verbose, benchmarkOnly=True)

            plt = importMatplotlib()
            print("  > Plotting %d figures." %(len(fastmat.core.calData)))

            def plot(arrN, arrTime, arrNested, arrEstimate,
                     title, legend=False):
                '''
                Plot a time measurement and -model for each a Forward() and
                Backward() transform
                '''
                plt.loglog(arrN, arrTime[:, 0],
                           color=colors[0], linestyle='dotted')
                plt.loglog(arrN, arrTime[:, 1],
                           color=colors[1], linestyle='dotted')
                plt.loglog(arrN, arrNested[:, 0],
                           color=colors[0], linestyle='dashed')
                plt.loglog(arrN, arrNested[:, 1],
                           color=colors[1], linestyle='dashed')
                plt.loglog(arrN, arrEstimate[:, 0],
                           color=colors[0], linestyle='solid')
                plt.loglog(arrN, arrEstimate[:, 1],
                           color=colors[1], linestyle='solid')
                if legend:
                    plt.legend(['%s-%s' %(a, b)
                                for a in ['time measure', 'nested matrices',
                                          'estimated model']
                                for b in ['Forward()', 'Backward()']])
                plt.title(title)

            bb = 1
            numBlocks = int(np.ceil(np.sqrt(len(benchmarks))))
            for item, bench in benchmarks.items():
                # arrN holds the problem size of each benchmark
                # arrTime holds the measured durations of forward and backward
                # arrNested states the amount of time spent in nested classes
                # for the forward (col 0) and backward (col 1) transform
                arrN = bench.getResult('overhead', 'numN')

                arrTime = bench.getResult('overhead',
                                          'forwardMin',
                                          'backwardMin')
                arrNested = bench.getResult('overhead',
                                            BENCH.RESULT_OVH_NESTED_F,
                                            BENCH.RESULT_OVH_NESTED_B,
                                            BENCH.RESULT_EFF_NESTED_F,
                                            BENCH.RESULT_EFF_NESTED_B)
                arrNested = arrNested[:, 0:2] + arrNested[:, 2:4]

                # now load the calibration data and generate the estimate based
                # on the calibration and the transforms' complexity estimates
                cal = fastmat.core.getMatrixCalibration(item)
                matCal = np.diag(np.array([cal.gainForward, cal.gainBackward]))
                arrComplexity = bench.getResult('overhead',
                                                BENCH.RESULT_COMPLEXITY_F,
                                                BENCH.RESULT_COMPLEXITY_B)
                arrEstimate = (
                    np.array([[cal.offsetForward, cal.offsetBackward]]) +
                    arrNested + arrComplexity.dot(matCal))

                # final step: plotting into a new subplot each
                plt.subplot(numBlocks, numBlocks, bb)
                bb = bb + 1
                plot(arrN, arrTime, arrNested, arrEstimate, item.__name__)

            # plot an empty diagram with the same line parameters as the other
            # plots, but this time with a legend -> generates a legend frame
            plt.subplot(numBlocks, numBlocks, bb)
            plot(np.empty((0, 2)), np.empty((0, 2)),
                 np.empty((0, 2)), np.empty((0, 2)), 'Legend', legend=True)

            plt.show()

    ############################################## command: performance
    def performance(self, *arguments):
        '''Extract built-in documentation from units in package.'''
        self.argParser.add_argument(
            'classes',
            nargs='*',
            default=[],
            help="Select the classes to show in performance plots. " +
            "`all` is default."
        )
        self.argParser.add_argument(
            '-c', '--calibration',
            type=str,
            default='',
            help="Filename to read calibration data from. Loads calibration " +
            "data for fastmat classes which is then used by the package."
        )

        # parse arguments
        self.args = self.argParser.parse_args(arguments)

        # determine a list of selected classes: interpret codewords
        if len(self.args.classes) < 1:
            classes = fastmat.classes
        else:
            classNames = {item.__name__: item for item in fastmat.classes}
            classes = [classNames[name]
                       for name in self.args.classes
                       if name in classNames]

        # load calibration data from file if specified
        if len(self.args.calibration) > 0:
            fastmat.core.loadCalibration(self.args.calibration)
            print("  > Loaded calibration data from %s." %(
                self.args.calibration))

        # run the benchmarks and plot the result
        plt = importMatplotlib()

        def plot(arrN, arrTime, arrEstimate, title, legend=False):
            '''
            Plot a time measurement and -model for each a Forward() and
            Backward() transform
            '''
            plotLegend = ['time measure']
            plt.loglog(arrN, arrTime[:, 0],
                       color=colors[0], linestyle='dotted')
            plt.loglog(arrN, arrTime[:, 1],
                       color=colors[1], linestyle='dotted')
            if arrEstimate.shape[0] == arrN.shape[0]:
                plt.loglog(arrN, arrEstimate[:, 0],
                           color=colors[0], linestyle='solid')
                plt.loglog(arrN, arrEstimate[:, 1],
                           color=colors[1], linestyle='solid')
                plotLegend.append('estimated model')

            if legend:
                plt.legend(['%s-%s' %(a, b)
                            for a in plotLegend
                            for b in ['Forward()', 'Backward()']])
            plt.title(title)

        print("  > Generating performance measurements for class plots")
        bb = 1
        numBlocks = int(np.ceil(np.sqrt(len(classes) + 1)))
        for item in classes:
            print("     - %s" %(item.__name__))
            bench = fastmat.core.calibrateClass(item, benchmarkOnly=True)

            # arrN holds the problem size of each benchmark
            # arrTime holds the measured durations of forward and backward
            # arrNested states the amount of time spent in nested classes
            # for the forward (col 0) and backward (col 1) transform
            arrN = bench.getResult('overhead', 'numN').squeeze()

            arrTime = bench.getResult('overhead',
                                      'forwardMin',
                                      'backwardMin')

            # fetch the runtime estimation from the benchmarks
            arrEstimate = bench.getResult('overhead',
                                          BENCH.RESULT_ESTIMATE_FWD,
                                          BENCH.RESULT_ESTIMATE_BWD)

            # final step: plotting into a new subplot each
            plt.subplot(numBlocks, numBlocks, bb)
            bb = bb + 1
            plot(arrN, arrTime, arrEstimate, item.__name__)

        # plot an empty diagram with the same line parameters as the other
        # plots, but this time with a legend -> generates a legend frame
        plt.subplot(numBlocks, numBlocks, bb)
        plot(np.empty((0, 2)), np.empty((0, 2)), np.empty((0, 2)),
             'Legend', legend=True)

        print("  > Plotting %d figures." %(len(classes)))
        plt.show()


################################################################################
################################################## Script entry point
if __name__ == '__main__':
    ################################################## Collect information
    # keep track of already seen items during the search
    seenItems=set([])

    def crawlModule(module, level, targets):
        elements={}

        # determine elements of package
        items=getObjects(module)

        # determine type of package elements
        for itemName in items:
            # fetch element
            try:
                item=getattr(module, itemName)
            except AttributeError:
                continue

            # determine module name and package descendency
            # Also check, that the following conditions are met:
            #  -> the module is a part of our package
            #  -> we haven't seen it yet
            #  -> the except AttributeError catches errors caused by numeric
            #     function names sometimes generated by cython
            try:
                moduleName=item.__module__ \
                    if not inspect.ismodule(item) else item.__name__
                modulePath=moduleName.split('.')
            except AttributeError:
                continue

            try:
                if modulePath[0] != packageName or (item in seenItems):
                    continue
            except TypeError:
                continue

            seenItems.add(item)

            # if the element is a submodule of fastmat, go deeper
            if inspect.ismodule(item):
                result = crawlModule(item, level + 1, targets)
                if len(result) > 0:
                    elements[itemName] = result
            elif (inspect.isclass(item) and issubclass(item, targets)):
                elements[itemName] = item

        return elements

    # Index elements in fastmat by location
    packageIndexTree = crawlModule(fastmat, 1, classBaseContainers)

    ############################################## Collect package infos
    # Flatten Index for categorization
    def appendDictFlattened(
        output,
        prefix,
        dictionary,
        separator='.'
    ):
        '''Flatten dict, prefixing nested levels with its key value'''
        if len(prefix) > 0:
            prefix=prefix + separator

        # iterate elements, nest if dict was found, add otherwise
        for key, item in dictionary.items():
            if isinstance(item, dict):
                appendDictFlattened(output, prefix + key, item)
            else:
                output[prefix + key]=item

    packageIndex={}
    appendDictFlattened(packageIndex, '', packageIndexTree)

    # now filter out the fastmat.Algorithm class as it is just a classifier
    packageIndex={name: element
                  for name, element in packageIndex.items()
                  if element is not fastmat.Algorithm}

    ############################################## analyze structure
    # categorize into classes (class) and algorithms (routine)
    packageClasses={loc: element
                    for loc, element in packageIndex.items()
                    if issubclass(element, fastmat.Matrix)}

    packageAlgs={loc: element
                 for loc, element in packageIndex.items()
                 if issubclass(element, fastmat.Algorithm)}

    ################################################## Parse and Run
    Bee(*(sys.argv[1:]), prevArgs=' '.join(sys.argv[:1]))



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  demo/matrixSAFT.py
 -------------------------------------------------- part of the fastmat demos

  Demonstration on how to use fastMat for SAFT.


  Author      : wcw
  Introduced  :
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
# import numpy functionality
import numpy as np
cimport numpy as np

ctypedef fused DATATYPE:
    np.int8_t
    np.int16_t
    np.int32_t
    np.int64_t
    np.float32_t
    np.float64_t
    np.complex64_t
    np.complex128_t

###############################################################################
# cython-optimized fastmat matrix class core for efficient mask matrices
###############################################################################
cdef void _SaftMaskCore(
    DATATYPE[:, :] mvIn,
    DATATYPE[:, :] mvOut,
    int sizeItem,
    int numBlocks,
    tuple masks,
    bint backward
):
    cdef int mm, ee, ii, iiA, iiB, oo, vv, iiAe, iiBe, ooe, dC
    cdef int numElements, numMasks, numSize, numVecs
    cdef np.uint16_t[:] maskRow, maskCol
    cdef bint applyAbove

    cdef DATATYPE value

    numSize = mvIn.shape[0]
    numVecs = mvIn.shape[1]
    numMasks = len(masks)
    for mm in range(numMasks):
        # get mask and determine number of elements
        maskRow = masks[mm][:, 0 if backward else 1]
        maskCol = masks[mm][:, 1 if backward else 0]
        numElements = maskRow.shape[0]

        for bb in range(numBlocks):

            # compute elements on and below the main block diagonal
            # determine indices into input and output memoryviews
            oo = bb * sizeItem
            iiB = (bb - mm) * sizeItem
            iiA = (bb + mm) * sizeItem
            applyAbove = iiA < numSize and (iiA != iiB)

            for ee in range(numElements):
                ooe = oo + maskRow[ee]

                dC = maskCol[ee]
                iiAe = iiA + dC
                iiBe = iiB + dC

                for vv in range(numVecs):
                    value = mvIn[iiBe, vv] if iiBe >= 0 else 0
                    if applyAbove:
                        value += mvIn[iiAe, vv]
                    if value != 0:
                        mvOut[ooe, vv] = mvOut[ooe, vv] + value




################################################## type-dispatch
cpdef np.ndarray SaftMaskCore(
    np.ndarray arrIn,
    int sizeItem,
    int numBlocks,
    tuple masks,
    bint backward
):
    cdef np.ndarray arrOut = np.zeros((<object> arrIn).shape, dtype=arrIn.dtype)

    if arrIn.dtype == np.int8:
        _SaftMaskCore[np.int8_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    elif arrIn.dtype == np.int16:
        _SaftMaskCore[np.int16_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    elif arrIn.dtype == np.int32:
        _SaftMaskCore[np.int32_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    elif arrIn.dtype == np.int64:
        _SaftMaskCore[np.int64_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    elif arrIn.dtype == np.float32:
        _SaftMaskCore[np.float32_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    elif arrIn.dtype == np.float64:
        _SaftMaskCore[np.float64_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    elif arrIn.dtype == np.complex64:
        _SaftMaskCore[np.complex64_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    elif arrIn.dtype == np.complex128:
        _SaftMaskCore[np.complex128_t](
            arrIn, arrOut, sizeItem, numBlocks, masks, backward)
    else:
        raise TypeError("Data type '%s' not supported in SaftMask.forward()" %(
            str(arrIn.dtype)))

    return arrOut



# -*- coding: utf-8 -*-
#cython: boundscheck=False, wraparound=False
'''
  fastmat/algs/CG.pyx
 -------------------------------------------------- part of the fastmat package

  Implementation of conjugated gradient method solver for linear equation
  systems in fastmat.


  Author      : sempersn
  Introduced  : 2016-04-08
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner
       https://www.tu-ilmenau.de/ems/

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''
import numpy as np
cimport numpy as np

from ..core.cmath cimport _arrEmpty, _arrZero, _arrForceTypeAlignment, _norm
from ..core.types cimport *
from ..base import Algorithm

from ..Matrix cimport Matrix


################################################################################
###  CG: Conjugate gradient method solver for linear equation systems
################################################################################
cpdef np.ndarray CG(
    Matrix fmatA,
    np.ndarray arrB,
    float eps=0
):
    '''Solve linear equation system 'fmatA * x = arrB' for x.'''
    if arrB.ndim > 2:
        raise ValueError("Only N x M arrays are supported for CG")

    cdef np.dtype typeOut = np.promote_types(
        np.float32, np.promote_types(fmatA.dtype, arrB.dtype))
    cdef nptype npTypeOut = typeOut.type_num
    arrIn = _arrForceTypeAlignment(arrB, npTypeOut, np.NPY_FORCECAST)
    if eps == 0:
        eps = _getTypeEps(typeOut)

    # dispatch specialization of core routine according tOptor
    if typeOut == np.float32:
        return _CGcore[np.float32_t](fmatA, arrIn, npTypeOut, 0., eps)
    elif typeOut == np.float64:
        return _CGcore[np.float64_t](fmatA, arrIn, npTypeOut, 0., eps)
    elif typeOut == np.complex64:
        return _CGcore[np.complex64_t](fmatA, arrIn, npTypeOut, 0., eps)
    elif typeOut == np.complex128:
        return \
            _CGcore[np.complex128_t](fmatA, arrIn, npTypeOut, 0., eps)
    else:
        raise NotImplementedError("Output type %d not supported." % (typeOut))


cdef np.ndarray _CGcore(
    Matrix fmatA,
    np.ndarray arrB,
    nptype npTypeOut,
    TYPE_FLOAT typeTag,
    float eps
):
    '''
    Solve linear equation system 'fmatA * x = arrB' for x.
    The following variables are used:
    arrB         - input data array
    fmatA        - input system matrix
    arrR         - residual vector (TYPE_FLOAT, CONT)
    vecR         - residual vector (TYPE_FLOAT[:] in arrR)
    arrP         - next krylov subspace vector (TYPE_FLOAT[:N], CONT)
    vecP         - next krylov subspace vector (TYPE_FLOAT[:] in arrP)
    vecQ         - projection of krylov subspace vector onto cols of matA
    arrIn        - symmetrized right-hand side [TYPE_FLOAT, F, CONT]
    arrOut       - current solution [TYPE_FLOAT, F, CONT]
    vecOut       - vector of current solution (TYPE_FLOAT[:] in arrOut)
    numAlpha     - optimal step with
    numRNormNew  - new residual norm
    numRNormOld  - old residual norm
    eps          - stopping condition to the projected residual
    '''
    # NOTE: typeTag is used for telling the compiler the used specialization
    # fetch dimensions of arrB
    cdef intsize mm, nn, numStep
    cdef intsize N = arrB.shape[0]
    cdef intsize M = arrB.shape[1] if arrB.ndim > 1 else 1

    # allocate temporary variables used in vector loop
    cdef np.float64_t numRNormOld, numRNormNew, ratio
    cdef TYPE_FLOAT numValue, numCorr, numAlpha

    # allocate temporary buffers in memory and create memoryviews into them
    cdef np.ndarray arrOut, arrP = _arrEmpty(1, N, 1, npTypeOut)
    cdef TYPE_FLOAT * vecP = <TYPE_FLOAT * > arrP.data
    cdef np.ndarray arrQ
    cdef TYPE_FLOAT * vecQ

    # change right hand side of equation system according to symmetrization
    # force to be F-contiguous and of consistent data type (no ints here)
    cdef np.ndarray arrR = _arrForceTypeAlignment(
        fmatA.backward(arrB), npTypeOut, np.NPY_FORCECAST)
    cdef TYPE_FLOAT * pArrR = <TYPE_FLOAT * > arrR.data
    cdef TYPE_FLOAT * vecR
    arrOut = _arrZero(2, fmatA.numM, M, npTypeOut)
    cdef TYPE_FLOAT * pArrOut = <TYPE_FLOAT * > arrOut.data
    cdef TYPE_FLOAT * vecOut

    # iterate over vectors
    for mm in range(M):

        vecOut = &(pArrOut[mm * N])
        vecR   = &(pArrR[mm * N])

        # norm of residual
        numRNormOld = _norm(vecR, N)

        # projected residual (initialized with RHS-reformulation),
        # use to bootstrap krylov subspace vector by copy
        for nn in range(N):
            vecP[nn] = vecR[nn]

        # step counter
        numStep = 0

        # iterate until stopping criterion is met
        while numRNormOld > eps:
            arrQ = fmatA.gram.forward(arrP)
            vecQ = <TYPE_FLOAT * > arrQ.data

            # calculate next optimal step width according to
            # current residual and correlation between old and
            # new projection: correlate np.conj(arrP) (*) arrQ
            numCorr = 0
            for nn in range(N):
                numValue = vecP[nn]
                # conjugate elements of P on-the-fly
                if (TYPE_FLOAT == np.complex64_t) or \
                        (TYPE_FLOAT == np.complex128_t):
                    numValue.imag = -numValue.imag

                numCorr += numValue * vecQ[nn]

            numAlpha = numRNormOld / numCorr

            # update current solution and residual
            # here we use the orthogonality of arrP and arrQ
            # with respect to the inner product induced by A
            # while iterating, calculate norm of vecR on-the-fly
            for nn in range(N):
                vecOut[nn] = vecOut[nn] + numAlpha * vecP[nn]
                vecR[nn] = vecR[nn] - numAlpha * vecQ[nn]

            # new residual norm
            numRNormNew = _norm(vecR, N)

            # next krylov subspace vector
            value = numRNormNew / numRNormOld
            for nn in range(N):
                vecP[nn] = vecP[nn] * value + vecR[nn]

            numRNormOld = numRNormNew

            # step increment
            numStep += 1

    return arrOut


################################################################################
###  Maintenance and Documentation
################################################################################

################################################## inspection interface
class CGinspect(Algorithm):

    def _getTest(self):
        from ..inspect import TEST, dynFormat, arrTestDist

        def testCG(test):

            # prepare vectors
            test[TEST.RESULT_REF]    = arrTestDist((test[TEST.NUM_M],
                                                    test[TEST.DATACOLS]),
                                                   dtype=test[TEST.DATATYPE])
            test[TEST.RESULT_INPUT]  = (test[TEST.INSTANCE] *
                                        test[TEST.RESULT_REF])
            test[TEST.RESULT_OUTPUT] = CG(test[TEST.INSTANCE],
                                          test[TEST.RESULT_INPUT])

        return {
            TEST.ALGORITHM: {
                TEST.NUM_N      : 27,
                TEST.NUM_M      : TEST.NUM_N,

                'typeA'         : TEST.Permutation(TEST.ALLTYPES),
                'arrA'          : TEST.ArrayGenerator({
                    TEST.DTYPE  : 'typeA',
                    TEST.SHAPE  : (TEST.NUM_N, TEST.NUM_N)
                }),

                TEST.OBJECT     : Matrix,
                TEST.INITARGS   : (lambda param: [param.arrA()]),

                TEST.DATAALIGN  : TEST.ALIGNMENT.DONTCARE,
                TEST.INIT_VARIANT : TEST.IgnoreFunc(testCG),

                'strTypeA'      : (lambda param: TEST.TYPENAME[param['typeA']]),
                TEST.NAMINGARGS : dynFormat("%s", 'arrA'),

                # matrix inversion always expands data type to floating-point
                TEST.TYPE_PROMOTION : np.float32,
                #TEST.CHECK_PROXIMITY : False,
                TEST.TOL_POWER  : 7.
            },
        }

    def _getBenchmark(self):
        from ..inspect import BENCH, arrTestDist
        from ..Diag import Diag

        def createTarget(M, datatype):
            '''Create test target for algorithm performance evaluation.'''

            # generate matA (random diagonal matrix)
            matA = Diag(arrTestDist((M, 1), datatype))

            # generate arrb from random baseline support and matrix (RHS)
            arrB = matA.forward(arrTestDist((M, 1), datatype))

            return (CG, [matA, arrB])

        return {
            BENCH.COMMON: {
                BENCH.NAME      : 'Method of Conjugate Gradients',
                BENCH.DOCU      : r'$\bm A = \diag(\{1,\dots,n\})$',
                BENCH.FUNC_GEN  : (lambda c: createTarget(c, np.double))
            },
            BENCH.PERFORMANCE: {
                BENCH.CAPTION   : 'CG performance'
            },
            BENCH.DTYPES: {
                BENCH.FUNC_GEN  : createTarget,
                BENCH.FUNC_SIZE : (lambda c: c),
                BENCH.FUNC_STEP : (lambda c: c * 10 ** (1. / 12)),
            }
        }

    def _getDocumentation(self):
        from ..inspect import DOC
        return DOC.SUBSECTION(
            r'Conjugate Gradient Method (CG) (\texttt{fastmat.algs.CG})',
            DOC.SUBSUBSECTION(
                'Definition and Interface', r"""
For a given full rank hermitian matrix $\bm A \in \C^n$ and a vector $\bm b \in
\C^n$, we solve

\begin{align}
\bm A \cdot \bm x = \bm b
\end{align}

for $\bm x \in \C^n$, i.e. $\bm x = \bm A^{-1} \cdot \bm b$. If $\bm A$ is not
hermitian, we solve

\begin{align}
\bm A^\herm \cdot \bm A \cdot \bm x = \bm A^\herm \cdot \bm b
\end{align}
instead. In this case it should be noted, that the condition number of $\bm
A^\herm \cdot \bm A$ might be a lot larger than the one of $\bm A$ an thus we
might run into stability problems for large and already ill-conditioned systems.

\begin{itemize}
\item \textbf{Input:} System matrix $\bm A$, right hand side $\bm b$ and
stopping tolerance for the residual $0 < \varepsilon \ll 1$.
\item \textbf{Output:} Solution $\bm x = \bm A^{-1} \cdot \bm b$
\end{itemize}

This algorithm was originally described in \cite{cg_stiefel1952cg} and is
applicable here, because it only uses the backward and forward projection of a
matrix.""",
                DOC.SNIPPET('# import the packages',
                            'import numpy.random as npr',
                            'import numpy as np',
                            'import fastmat as fm',
                            'import fastmat.algs as fma',
                            '',
                            '# construct the matrix',
                            'n = 26',
                            'H = fm.Hadamard(n)',
                            '',
                            '# define the right hand side',
                            'b = npr.randn(2 ** n)',
                            '',
                            '# solve the system',
                            'y = fma.CG(H, b)',
                            '',
                            '# check if solution is correct',
                            'print(np.allclose(',
                            '    b, H.forward(y)))',
                            caption=r"""
We construct a Hadamard matrix of order $26$, which would consume
\SI{4.5}{\peta\byte} of memory if we used \SI{1}{byte} integers to represent it
and solve above system of linear equations.""")
            ),
            DOC.SUBSUBSECTION(
                'Performance Benchmarks', r"""
All benchmarks were performed on a matrix
$\bm \Hs_k$ and $n = 2^k$ with $n, k in \N$""",
                DOC.PLOTPERFORMANCE(),
                DOC.PLOTTYPESPEED(),
                DOC.PLOTTYPEMEMORY()
            ),
            DOC.BIBLIO(
                cg_stiefel1952cg=DOC.BIBITEM(
                    r'Hestenes, Magnus R., Stiefel, Eduard',
                    r"""
Methods of Conjugate Gradients for Solving Linear Systems""",
                    r"""
The Springer International Series in Engineering and Computer Science,
Volume 383, 1997""")
            )
        )



# -*- coding: utf-8 -*-
'''
  fastmat/__init__.py
 -------------------------------------------------- part of the fastmat demos


  Author      : wcw
  Introduced  :
 ------------------------------------------------------------------------------

   Copyright 2016 Sebastian Semper, Christoph Wagner

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

 ------------------------------------------------------------------------------
'''

# import fundamental types and classes first, also behavioural flags
from .Matrix import Matrix, Hermitian, Conjugate, Transpose, flags
from .Product import Product

# import all fastmat classes
from .BlockDiag import BlockDiag
from .Blocks import Blocks
from .Circulant import Circulant
from .Diag import Diag
from .Eye import Eye
from .Fourier import Fourier
from .Hadamard import Hadamard
from .Kron import Kron
from .LowRank import LowRank
from .Outer import Outer
from .Parametric import Parametric
from .Partial import Partial
from .Permutation import Permutation
from .Polynomial import Polynomial
from .Sparse import Sparse
from .Sum import Sum
from .Toeplitz import Toeplitz
from .Zero import Zero

# new stuff
from .DiagBlocks import DiagBlocks
from .LFSRCirculant import LFSRCirculant
from .MLCirculant import MLCirculant
from .MLToeplitz import MLToeplitz
from .MLUltraSound import MLUltraSound

# import algorithms subpackage
from .base import Algorithm
from . import algs


# define package version (gets overwritten by setup script)
from .version import __version__

# compile a list of all matrix classes offered by the package


def isClass(item):
    try:
        return issubclass(item, Matrix)
    except TypeError:
        return False


classes = list(item for item in locals().values() if isClass(item))
